import asyncio

import aiorpcx
from ..transaction import *
from ..utils import is_hex
from binascii import unhexlify
from ..blocks import verify_merkle_proof, deserialize_header
from .. import segwit_addr
from ..electrumx_client import ElectrumXClient
from ..keystore import *
from ..wallet import *
from ..py3specials import *
from ..constants import SATOSHI_PER_BTC
from ..opcodes import opcodes
from functools import partial
from typing import Dict, Any, Tuple, Optional, Union, Iterable, Type, Callable, Generator, AsyncGenerator
from ..types import (Tx, Witness, TxInput, TxOut, BlockHeader, MerkleProof, AddressBalance, BlockHeaderCallback,
                     AddressCallback, AddressTXCallback, PrivkeyType, PrivateKeySignAllType, TXInspectType, PubKeyType)
from ..electrumx_client.types import (ElectrumXBlockHeaderNotification, ElectrumXHistoryResponse,
                                      ElectrumXBalanceResponse, ElectrumXUnspentResponse, ElectrumXTx,
                                      ElectrumXMerkleResponse, ElectrumXMultiBalanceResponse, ElectrumXMultiTxResponse,
                                      ElectrumXVerboseTX)
from cryptos.utils import alist


class TXInvalidError(BaseException):
    pass


class TXRejectedError(TXInvalidError):
    pass


class BaseCoin:
    """
    Base implementation of crypto coin class
    All child coins_async must follow same pattern.
    """

    coin_symbol: str = None
    display_name: str = None
    enabled: bool = True
    segwit_supported: bool = None
    cash_address_supported: bool = False
    magicbyte: int = None
    script_magicbyte: int = None
    segwit_hrp: str = None
    cash_hrp: str = None
    explorer: Type[ElectrumXClient] = ElectrumXClient
    client_kwargs: Dict[str, Any] = {
        'server_file': 'bitcoin.json',
        'use_ssl': True
    }
    _client: ElectrumXClient = None
    _block: Tuple[int, str, BlockHeader] = None
    is_testnet: bool = False
    testnet_overrides: Dict[str, Any] = {}
    hashcode: int = SIGHASH_ALL
    secondary_hashcode: Optional[int] = None
    hd_path: int = 0
    block_interval: int = 10
    minimum_fee: int = 500
    txid_bytes_len = 32
    signature_sizes: Dict[str, int] = {
        'p2pkh': 213,
        'p2w_p2sh': 46 + (213 / 4),
        'p2wpkh': (214 / 4),
    }
    wif_prefix: int = 0x80
    wif_script_types: Dict[str, int] = {
        'p2pkh': 0,
        'p2wpkh': 1,
        'p2wpkh-p2sh': 2,
        'p2sh': 5,
        'p2wsh': 6,
        'p2wsh-p2sh': 7
    }
    xprv_headers: Dict[str, int] = {
        'p2pkh': 0x0488ade4,
        'p2wpkh-p2sh': 0x049d7878,
        'p2wsh-p2sh': 0x295b005,
        'p2wpkh': 0x4b2430c,
        'p2wsh': 0x2aa7a99
    }
    xpub_headers: Dict[str, int] = {
        'p2pkh': 0x0488b21e,
        'p2wpkh-p2sh': 0x049d7cb2,
        'p2wsh-p2sh': 0x295b43f,
        'p2wpkh': 0x4b24746,
        'p2wsh': 0x2aa7ed3
    }
    electrum_xprv_headers: Dict[str, int] = xprv_headers
    electrum_xpub_headers: Dict[str, int] = xpub_headers

    def __init__(self, testnet: bool = False, **kwargs):
        if testnet:
            self.is_testnet = True
            for k, v in self.testnet_overrides.items():
                setattr(self, k, v)
        # override default attributes from kwargs
        for key, value in kwargs.items():
            if isinstance(value, dict):
                getattr(self, key).update(value)
            else:
                setattr(self, key, value)
        if not self.enabled:
            if self.is_testnet:
                raise NotImplementedError(
                    f"Due to explorer limitations, testnet support for {self.display_name} has not been implemented yet!")
            else:
                raise NotImplementedError(f"Support for {self.display_name} has not been implemented yet!")
        self.address_prefixes: List[str] = magicbyte_to_prefix(magicbyte=self.magicbyte)
        self.script_prefixes: List[str] = []
        if self.script_magicbyte:
            self.script_prefixes = magicbyte_to_prefix(magicbyte=self.script_magicbyte)
        self.secondary_hashcode = self.secondary_hashcode or self.hashcode
        self.fees = {}

    @property
    def client(self):
        """
        Connect to remote server
        """
        if not self._client:
            self._client = self.explorer(**self.client_kwargs)
        return self._client

    async def close(self) -> None:
        if self._client:
            await self._client.close()

    async def estimate_fee_per_kb(self, numblocks: int = 6) -> float:
        """
        Get estimated fee kb to get transaction confirmed within numblocks number of blocks
        """
        return await self.client.estimate_fee(numblocks)

    def tx_size(self, txobj: Tx) -> float:
        """
        Get transaction size in bytes
        """
        tx = serialize(txobj)
        size = len(tx) / 2
        for inp in txobj['ins']:
            address = inp.get('address')
            if address and self.is_native_segwit(address):
                pass        # Segwit signatures not included in tx size for fee purposes?
            elif address and self.maybe_legacy_segwit(address):
                size += self.signature_sizes['p2w_p2sh']    # Not sure if segwit or not
            else:
                size += self.signature_sizes['p2pkh']
        return size

    async def estimate_fee(self, txobj: Tx, numblocks: int = 6) -> int:
        """
        Get estimated fee to get transaction confirmed within numblocks number of blocks.
        txobj is a pre-signed transaction object
        """
        num_bytes = self.tx_size(txobj)
        btc_fee_per_kb = await self.estimate_fee_per_kb(numblocks=numblocks)
        if btc_fee_per_kb > 0:
            btc_fee_per_byte = btc_fee_per_kb / 1024
            satoshi_fee_per_byte = btc_fee_per_byte * SATOSHI_PER_BTC
            return int(num_bytes * satoshi_fee_per_byte)
        return 0

    @staticmethod
    async def _tasks_with_inputs(coro: Callable, *args: Any, **kwargs) -> Generator[Tuple[str, Any], None, None]:
        for i, result in enumerate(await asyncio.gather(*[coro(arg, **kwargs) for arg in args])):
            arg = args[i]
            yield arg, result

    async def raw_block_header(self, height: int) -> str:
        return await self.client.block_header(height)

    async def block_header(self, height: int) -> BlockHeader:
        """
        Return block header data for the given height
        """
        header = await self.raw_block_header(height)
        return deserialize_header(unhexlify(header))

    async def block_headers(self, *args: int) -> AsyncGenerator[BlockHeader, None]:
        """
        Return block header data for the given heights
        """
        for header in await asyncio.gather(*[self.block_header(h) for h in args]):
            yield header

    @staticmethod
    def _get_block_header_notification_params(header: ElectrumXBlockHeaderNotification) -> Tuple[int, str, BlockHeader]:
        height = header['height']
        hex_header = header['hex']
        header = deserialize_header(unhexlify(hex_header))
        return height, hex_header, header

    @staticmethod
    async def _await_or_in_executor(func: Callable, *args):
        f = func
        is_coro = asyncio.iscoroutinefunction(f)
        while not is_coro and isinstance(f, partial):
            f = f.func
            is_coro = asyncio.iscoroutinefunction(f)
        if is_coro:
            await func(*args)
        else:
            try:
                await asyncio.get_running_loop().run_in_executor(None, func, *args)
            except RuntimeError as e:
                if "Non-thread-safe" in e:
                    """
                    Sync callbacks are called in another thread so cannot interact with asyncio objects.
                    """
                    raise Exception(
                        "Syncronous callbacks cannot interact with asyncio objects such as Futures. Make the callback a coroutine function.")
                else:
                    raise e

    async def _block_header_callback(self, callback: BlockHeaderCallback,
                                     header: ElectrumXBlockHeaderNotification) -> None:
        height, hex_header, header = self._get_block_header_notification_params(header)
        await self._await_or_in_executor(callback, height, hex_header, header)

    async def subscribe_to_block_headers(self, callback: BlockHeaderCallback) -> None:
        """
        Run callback when a new block is added to the blockchain
        Callback should be in the format:

        from cryptos.types import BlockHeader

        async def on_block_headers(height: int, hex_header: str, header:BlockHeader) -> None:
            pass

        or

        def on_block_headers(height: int, hex_header: str, header:BlockHeader) -> None:
            pass

        """

        callback = partial(self._block_header_callback, callback)
        return await self.client.subscribe_to_block_headers(callback)

    async def unsubscribe_from_block_headers(self) -> None:
        """
        Unsubscribe from running callbacks when a new block is added
        """
        return await self.client.unsubscribe_from_block_headers()

    async def _update_block(self, fut: asyncio.Future, height: int, hex_header: str, header: BlockHeader):
        self._block = (height, hex_header, header)
        if not fut.done():
            fut.set_result(True)

    @property
    async def block(self) -> Tuple[int, str, BlockHeader]:
        """
        Gets the latest block in the blockchain.
        First time this is run it will subscribe to block headers.
        """
        if not self._block:
            fut = asyncio.Future()
            await self.subscribe_to_block_headers(partial(self._update_block, fut))
            await fut
        return self._block

    def is_closing(self) -> bool:
        return not self.client or self.client.is_closing

    async def confirmations(self, height: int) -> int:
        if height > 0:
            return (await self.block)[0] - height + 1
        return 0

    async def _address_status_callback(self, callback: AddressCallback, address: str,
                                       scripthash: str, status: str) -> None:
        await self._await_or_in_executor(callback, address, status)

    async def subscribe_to_address(self, callback: AddressCallback, addr: str) -> None:
        """
        Run callback when an address changes (e.g. a new transaction)
        Callback should be in the format:

        def on_address_event(address: str, status: str) -> None:
            pass
        """

        orig_addr = addr
        if self.client.requires_scripthash:
            addr = self.addrtoscripthash(addr)
        callback = partial(self._address_status_callback, callback, orig_addr)
        return await self.client.subscribe_to_address(callback, addr)

    async def unsubscribe_from_address(self, addr: str):
        """
        Unsubscribe from running callbacks when an address changes
        """
        if self.client.requires_scripthash:
            addr = self.addrtoscripthash(addr)
        return await self.client.unsubscribe_from_address(addr)

    async def _address_transaction_callback(self, callback: AddressTXCallback, history: ElectrumXHistoryResponse,
                                            address: str, status: str) -> None:
        updated_history, unspents, balance, merkle_proven = await asyncio.gather(self.history(address),
                                                                                 self.unspent(address),
                                                                                 self.get_balance(address),
                                                                                 self.balance_merkle_proven(address))
        if history == ["-"]:
            # First response
            new_txs = []
            history.clear()
            history += updated_history
            newly_confirmed = []
        else:
            tx_hashes = {t['tx_hash']: t['height'] for t in history}
            new_txs = [t for t in updated_history if t['tx_hash'] not in tx_hashes.keys()]
            newly_confirmed = [t for t in updated_history if
                               tx_hashes.get('tx_hash') and not t['height'] == tx_hashes['tx_hash']]

        prev_history = [tx for tx in history if tx not in newly_confirmed]
        for tx in newly_confirmed:
            history.remove(tx)
        history += new_txs
        await self._await_or_in_executor(callback, address, new_txs, newly_confirmed, prev_history, unspents,
                                         balance['confirmed'], balance['unconfirmed'], merkle_proven)

    async def subscribe_to_address_transactions(self, callback: AddressTXCallback, addr: str) -> None:
        """
        When an address changes retrieve transactions and balances and run a callback
        Callback should be in the format:

        def on_address_change(address: str, txs: List[ElectrumXTx], newly_confirmed: List[ElectrumXTx], history: List[ElectrumXTx], unspent: List[ElectrumTX], confirmed: int, unconfirmed: int, proven: int) -> None:
            pass

        Any transactions since the last notification are in Txs. All previous transactions are in history.
        Balances according to the network are in confirmed and unconfirmed.
        Balances confirmed locally is in proven.
        """

        history = ["-"]

        callback = partial(self._address_transaction_callback, callback, history)

        return await self.subscribe_to_address(callback, addr)

    async def get_balance(self, addr: str) -> ElectrumXBalanceResponse:
        """
        Get address balance
        """
        if self.client.requires_scripthash:
            addr = self.addrtoscripthash(addr)
        return await self.client.get_balance(addr)

    async def get_balances(self, *args: str) -> AsyncGenerator[ElectrumXMultiBalanceResponse, None]:
        async for addr, result in self._tasks_with_inputs(self.get_balance, *args):
            result['address'] = addr
            yield result

    async def get_merkle(self, tx: ElectrumXTx) -> Optional[ElectrumXMerkleResponse]:
        return await self.client.get_merkle(tx['tx_hash'], tx['height'])

    async def merkle_prove(self, tx: ElectrumXTx) -> MerkleProof:
        """
        Prove that information returned from server about a transaction in the blockchain is valid. Only run on a
        tx with at least 1 confirmation.
        """

        merkle, block_header = await asyncio.gather(self.get_merkle(tx), self.block_header(tx['height']))
        if not merkle:
            # Can happen if request is run immediately after pushing a transaction
            return {
                "tx_hash": tx['tx_hash'],
                'proven': False
            }
        proof = verify_merkle_proof(tx['tx_hash'], block_header['merkle_root'], merkle['merkle'], merkle['pos'])
        return proof

    async def merkle_prove_by_txid(self, tx_hash: str) -> MerkleProof:
        tx = await self.get_tx(tx_hash)
        return await self.merkle_prove(tx)

    async def _filter_by_proof(self, *txs: ElectrumXTx) -> Iterable[ElectrumXTx]:
        """
        Return only transactions with verified merkle proof
        """
        results = await asyncio.gather(*[self.merkle_prove(tx) for tx in txs])
        proven = [r['tx_hash'] for r in results if r['proven']]
        return filter(lambda tx: tx['tx_hash'] in proven, txs)

    async def unspent(self, addr: str, merkle_proof: bool = False) -> ElectrumXUnspentResponse:
        """
        Get unspent transactions for address
        """
        value = addr
        if self.client.requires_scripthash:
            value = self.addrtoscripthash(value)
        unspents = await self.client.unspent(value)
        for u in unspents:
            u['address'] = addr
        if merkle_proof:
            return list(await self._filter_by_proof(*unspents))
        return unspents

    async def get_unspents(self, *args: str, merkle_proof: bool = False
                           ) -> AsyncGenerator[ElectrumXMultiTxResponse, None]:
        async for addr, result in self._tasks_with_inputs(self.unspent, *args, merkle_proof=merkle_proof):
            for tx in result:
                tx['address'] = addr
                yield tx

    async def balance_merkle_proven(self, addr: str) -> int:
        result = sum(u['value'] for u in await self.unspent(addr, merkle_proof=True))
        return result

    async def balances_merkle_proven(self, *args: str) -> AsyncGenerator[AddressBalance, None]:
        async for addr, result in self._tasks_with_inputs(self.unspent, *args, merkle_proof=True):
            yield {'address': addr, 'balance': sum(tx['value'] for tx in result)}

    async def history(self, addr: str, merkle_proof: bool = False) -> ElectrumXHistoryResponse:
        """
        Get transaction history for address
        """
        if self.client.requires_scripthash:
            addr = self.addrtoscripthash(addr)
        txs = await self.client.get_history(addr)
        if merkle_proof:
            return list(await self._filter_by_proof(*txs))
        return txs

    async def get_histories(self, *args: str, merkle_proof: bool = False) -> AsyncGenerator[
                           ElectrumXMultiTxResponse, None]:
        async for addr, result in self._tasks_with_inputs(self.history, *args, merkle_proof=merkle_proof):
            for tx in result:
                tx['address'] = addr
                yield tx

    async def get_raw_tx(self, tx_hash: str) -> str:
        """
        Fetch transaction from the blockchain
        """
        return await self.client.get_tx(tx_hash)

    async def get_tx(self, tx_hash: str) -> Tx:
        """
        Fetch transaction from the blockchain and deserialise it to a dictionary
        """
        tx = await self.get_raw_tx(tx_hash)
        deserialized_tx = deserialize(tx)
        return deserialized_tx

    async def get_verbose_tx(self, tx_hash: str) -> ElectrumXVerboseTX:
        """
        Fetch transaction from the blockchain in verbose form
        """
        return await self.client.get_tx(tx_hash, verbose=True)

    async def get_txs(self, *args: str) -> AsyncGenerator[Tx, None]:
        for tx in await asyncio.gather(*[self.get_tx(tx_hash) for tx_hash in args]):
            yield tx

    async def ensure_values(self, tx: Tx) -> Tx:
        if not all(inp.get('value') for inp in tx['ins']):
            tx_hashes = list(dict.fromkeys([inp['tx_hash'] for inp in tx['ins'] if not inp.get('value')]))
            try:
                txs = await alist(self.get_txs(*tx_hashes))
            except StopIteration as e:  # Not sure what causes this intermittent error
                txs = []
            for inp in tx['ins']:
                pos = inp['tx_pos']
                prev_tx = next(filter(lambda t: txhash(serialize(t)) == inp['tx_hash'], txs))
                inp['value'] = prev_tx['outs'][pos]['value']
        return tx

    async def calculate_fee(self, tx: Tx) -> int:
        try:
            tx = await self.ensure_values(tx)
        except StopIteration as e:  # Not sure what causes this intermittent error
            pass
        in_value = sum(i['value'] for i in tx['ins'])
        out_value = sum(o['value'] for o in tx['outs'])
        return in_value - out_value

    async def pushtx(self, tx: Union[str, Tx]):
        """
        Push/ Broadcast a transaction to the blockchain
        """
        if not isinstance(tx, str):
            tx = serialize(tx)
        try:
            result = await self.client.broadcast_tx(tx)
            return result
        except (aiorpcx.jsonrpc.ProtocolError, aiorpcx.jsonrpc.RPCError) as e:
            tx_obj = deserialize(tx)
            message = f'{tx_obj}\n{tx}\n{e.message}'
            if any(code == e.code for code in (1, -32600)):
                if "fee" in e.message:
                    try:
                        message += f"Fee is {await self.calculate_fee(tx_obj)}"
                    except StopIteration:
                        pass
                raise TXRejectedError(message)
            raise TXInvalidError(message)

    def privtopub(self, privkey: PrivkeyType) -> str:
        """
        Get public key from private key
        """
        return privtopub(privkey)

    def pubtoaddr(self, pubkey: PubKeyType) -> str:
        """
        Get address from a public key
        """
        return pubtoaddr(pubkey, magicbyte=self.magicbyte)

    def get_address_variations(self, address: str) -> List[str]:
        if self.cash_address_supported:
            if self.is_cash_address(address):
                return [address, self.cash_address_to_legacy_addr(address)]
            else:
                return [self.legacy_addr_to_cash_address(address), address]
        else:
            return [address]

    def pub_is_for_p2pkh_addr(self, pubkey: PubKeyType, address: str) -> bool:
        return self.pubtoaddr(pubkey) == address or (
                self.cash_address_supported and self.pub_to_cash_address(pubkey) == address)

    def wiftoaddr(self, privkey: PrivkeyType) -> str:
        magicbyte, priv = b58check_to_bin(privkey)
        wif_magicbyte = magicbyte - self.wif_prefix
        if self.wif_script_types['p2pkh'] == wif_magicbyte:
            return self.privtop2pkh(privkey)
        if self.wif_script_types['p2wpkh-p2sh'] == wif_magicbyte:
            return self.privtop2wpkh_p2sh(privkey)
        if self.wif_script_types['p2wpkh'] == wif_magicbyte:
            return self.privtosegwitaddress(privkey)
        else:
            raise Exception("Address type for this wif-encoded private key not supported yet")

    def privtop2pkh(self, privkey: PrivkeyType) -> str:
        return privtoaddr(privkey, magicbyte=self.magicbyte)

    def privtoaddr(self, privkey: PrivkeyType) -> str:
        """
        Get address from a private key
        """
        privkey_format = get_privkey_format(privkey)
        if "wif" in privkey_format:
            return self.wiftoaddr(privkey)
        return self.privtop2pkh(privkey)

    def electrum_address(self, masterkey: AnyStr, n: int, for_change: int = 0) -> str:
        """
        For old electrum seeds
        """
        pubkey = electrum_pubkey(masterkey, n, for_change=for_change)
        return self.pubtoaddr(pubkey)

    def encode_privkey(self, privkey: PrivkeyType, formt: str, script_type: str = "p2pkh") -> PrivkeyType:
        return encode_privkey(privkey, formt=formt, vbyte=self.wif_prefix + self.wif_script_types[script_type])

    def is_p2pkh(self, addr: str) -> bool:
        """
        Legacy addresses only doesn't include Cash P2PKH Address
        """
        try:
            magicbyte, bin = b58check_to_bin(addr)
            return magicbyte == self.magicbyte
        except Exception:
            return False

    def is_cash_or_legacy_p2pkh_address(self, addr: str)-> bool:
        return self.is_p2pkh(addr) or self.is_cash_address(addr)

    def is_p2sh(self, addr: str) -> bool:
        """
        Check if addr is a pay to script address
        """
        try:
            magicbyte, bin = b58check_to_bin(addr)
            return magicbyte == self.script_magicbyte
        except Exception:
            return False

    def is_native_segwit(self, addr: str) -> bool:
        return self.segwit_supported and self.segwit_hrp and addr.startswith(self.segwit_hrp)

    def is_cash_address(self, addr: str) -> bool:
        return self.cash_address_supported and self.cash_hrp and addr.startswith(self.cash_hrp)

    def is_address(self, addr: str) -> bool:
        """
        Check if addr is a valid address for this chain
        """
        return self.is_p2pkh(addr) or self.is_p2sh(addr) or self.is_native_segwit(addr) or self.is_cash_address(addr) or is_public_key(addr)

    def maybe_legacy_segwit(self, addr: str) -> bool:
        if self.segwit_supported:
            script = self.addrtoscript(addr)
            return script.startswith(opcodes.OP_HASH160.hex() + '14') and script.endswith(opcodes.OP_EQUAL.hex())
        return False

    def is_p2wsh(self, addr: str) -> bool:
        return self.is_native_segwit(addr) and len(addr) == 62

    def is_segwit_or_p2sh(self, addr: str) -> bool:
        """
        Check if addr is a p2wpkh, p2wsh or p2sh script
        """
        return self.is_native_segwit(addr) or self.maybe_legacy_segwit(addr)

    def output_script_to_address(self, script: str) -> str:
        """
        Convert an output script to an address
        """
        segwit_hrp = self.segwit_hrp if self.segwit_supported else None
        cash_hrp = self.cash_hrp if self.cash_address_supported else None
        return output_script_to_address(script, self.magicbyte, self.script_magicbyte, segwit_hrp, cash_hrp)

    def scripttoaddr(self, script: str) -> str:
        """
        Convert an input public key hash to an address
        """
        if is_hex(script):
            script = binascii.unhexlify(script)
        # 0x14 is expected pubkey hash length
        pubkey_hash_prefix = binascii.unhexlify(opcodes.OP_DUP.hex() + opcodes.OP_HASH160.hex() + '14')
        pubkey_hash_suffix = binascii.unhexlify(opcodes.OP_EQUALVERIFY.hex() + opcodes.OP_CHECKSIG.hex())
        if script[:3] == pubkey_hash_prefix and script[-2:] == pubkey_hash_suffix and len(script) == 25:
            return bin_to_b58check(script[3:-2], self.magicbyte)  # pubkey hash address
        else:
            # BIP0016 scripthash address
            return bin_to_b58check(script[2:-1], self.script_magicbyte)

    def p2sh_scriptaddr(self, script: str) -> str:
        """
        Convert an output p2sh script to an address
        """
        if is_hex(script):
            script = binascii.unhexlify(script)
        return hex_to_b58check(hash160(script), self.script_magicbyte)

    def p2sh_segwit_addr(self, script: str) -> str:
        """
        Convert an output p2sh script to a Native Segwit P2WSH address
        """
        if is_hex(script):
            script = binascii.unhexlify(script)
        return self.scripthash_to_segwit_addr(bin_sha256(script))

    def scripthash_to_cash_addr(self, scripthash: bytes) -> str:
        return cashaddr.encode_full(self.cash_hrp, cashaddr.SCRIPT_TYPE, scripthash)

    def p2sh_cash_addr(self, script: str) -> str:
        """
        Convert an output p2sh script to a Bitcoin Cash address
        """
        if is_hex(script):
            script = binascii.unhexlify(script)
        return self.scripthash_to_cash_addr(bin_hash160(script))

    def addrtoscript(self, addr: str) -> str:
        """
        Convert an output address to a script
        """
        if self.is_native_segwit(addr):
            witver, witprog = segwit_addr.decode_segwit_address(self.segwit_hrp, addr)
            if witprog is not None:
                return mk_p2w_scripthash_script(witver, witprog)
        elif self.is_cash_address(addr):
            prefix, kind, hash_bin = cashaddr.decode(addr)
            hash_hex = safe_hexlify(hash_bin)
            if kind == 0:
                return mk_pubkey_script(hash_hex)
            return hash_to_scripthash_script(hash_hex)
        if self.is_p2sh(addr):
            return mk_scripthash_script(addr)
        elif self.is_p2pkh(addr):
            return addr_to_pubkey_script(addr)
        elif is_public_key(addr):
            return mk_p2pk_script(addr)
        raise Exception(f'Unrecognised address: {addr}')

    def addrtoscripthash(self, addr: str) -> str:
        """
        For electrumx requests
        """
        script = self.addrtoscript(addr)
        return script_to_scripthash(script)

    def pubtop2wpkh_p2sh(self, pub: str) -> str:
        """
        Convert a public key to a pay to witness public key hash address (P2WPKH-P2SH, required for segwit)
        """
        if not self.segwit_supported:
            raise Exception("Segwit not supported for this coin")
        if len(pub) > 70:
            pub = compress(pub)
        return self.scripttoaddr(mk_p2wpkh_script(pub))

    def privtop2wpkh_p2sh(self, priv: PrivkeyType) -> str:
        """
        Convert a private key to a pay to witness public key hash address
        """
        return self.pubtop2wpkh_p2sh(privtopub(priv))

    def hash_to_segwit_addr(self, pub_hash: AnyStr) -> str:
        """
        Convert a hash to the native segwit address format outlined in BIP-0173
        """
        if not self.segwit_supported:
            raise NotImplementedError(f"{self.display_name} does not support segwit")
        return segwit_addr.encode_segwit_address(self.segwit_hrp, 0, pub_hash)

    def scripthash_to_segwit_addr(self, script_hash: AnyStr) -> str:
        """
        Convert a script hash to the native segwit address format
        """
        if not self.segwit_supported:
            raise NotImplementedError(f"{self.display_name} does not support segwit")
        return segwit_addr.encode_segwit_address(self.segwit_hrp, 0, script_hash)

    def hash_to_cash_addr(self, pub_hash: AnyStr) -> str:
        """
        Convert a hash to a cash address
        """
        if not self.cash_address_supported:
            raise NotImplementedError(f"{self.display_name} does not support cash addresses")
        return cashaddr.encode_full(self.cash_hrp, cashaddr.PUBKEY_TYPE, pub_hash)

    def privtosegwitaddress(self, privkey: PrivkeyType) -> str:
        """
        Convert a private key to the new segwit address format outlined in BIP01743
        """
        return self.pub_to_segwit_address(self.privtopub(privkey))

    def pub_to_cash_address(self, pubkey: str) -> str:
        """
        Convert a public key to a cash address
        """
        return self.hash_to_cash_addr(pubkey_to_hash(pubkey))

    def privtocashaddress(self, privkey: PrivkeyType) -> str:
        """
        Convert a private key to a cash address
        """
        return self.pub_to_cash_address(self.privtopub(privkey))

    def legacy_addr_to_cash_address(self, addr: str) -> str:
        """
        Convert a legacy Bitcoin Address to a Bitcoin cash address
        """
        magicbyte, pubkey_hash = b58check_to_bin(addr)
        if magicbyte == self.magicbyte:
            return self.hash_to_cash_addr(pubkey_hash)
        elif magicbyte == self.script_magicbyte:
            return self.scripthash_to_cash_addr(pubkey_hash)
        else:
            raise Exception(f"Magic Byte {magicbyte} not recognised")

    def cash_address_to_legacy_addr(self, addr: str) -> str:
        """
        Convert a Bitcoin cash address to a legacy Bitcoin address
        """
        prefix, kind, pubkey_hash = cashaddr.decode(addr)
        if kind == 0:
            return bin_to_b58check(pubkey_hash, self.magicbyte)
        return bin_to_b58check(pubkey_hash, self.script_magicbyte)

    def pub_to_segwit_address(self, pubkey: str) -> str:
        """
        Convert a public key to the new segwit address format outlined in BIP01743
        """
        return self.hash_to_segwit_addr(pubkey_to_hash(compress(pubkey)))

    def mk_multsig_address(self, *args: str, num_required: int = None) -> Tuple[str, str]:
        """
        :param args: List of public keys to used to create multisig
        :param num_required: The number of signatures required to spend (defaults to number of public keys provided)
        :return: multisig script
        """
        num_required = num_required or len(args)
        script = mk_multisig_script(*args, num_required)
        address = self.p2sh_scriptaddr(script)
        return script, address

    def mk_multsig_segwit_address(self, *args: str, num_required: int = None) -> Tuple[str, str]:
        num_required = num_required or len(args)
        pubs = [compress(pub) for pub in args]
        script = mk_multisig_script(*pubs, num_required)
        address = self.p2sh_segwit_addr(script)
        return script, address

    def mk_multsig_cash_address(self, *args: str, num_required: int = None) -> Tuple[str, str]:
        num_required = num_required or len(args)
        script = mk_multisig_script(*args, num_required)
        address = self.p2sh_cash_addr(script)
        return script, address

    def apply_multisignatures(self, txobj: Tx, i: int, script, *args):
        inp = txobj['ins'][i]
        segwit = False
        try:
            if address := inp['address']:
                segwit = self.is_native_segwit(address)
        except (IndexError, KeyError):
            pass
        return apply_multisignatures(txobj, i, script, *args, segwit=segwit)

    def sign(self, txobj: Union[Tx, AnyStr], i: int, priv: PrivkeyType) -> Tx:
        """
        Sign a transaction input with index using a private key
        """
        if not isinstance(txobj, dict):
            txobj = deserialize(txobj)
        if len(priv) <= 33:
            priv = safe_hexlify(priv)
        pub = self.privtopub(priv)
        inp = txobj['ins'][i]
        p2pk = False
        segwit = False
        native_segwit = False
        try:
            if address := inp['address']:
                segwit = native_segwit = self.is_native_segwit(address)
                segwit = segwit or self.maybe_legacy_segwit(address)
                if segwit:
                    pub = compress(pub)
                elif len(address) in [66, 130]:
                   pub = address
                   p2pk = True
        except (IndexError, KeyError):
            pass
        if segwit:
            if 'witness' not in txobj.keys():
                txobj.update({"marker": 0, "flag": 1, "witness": []})
                for _ in range(0, i):
                    witness: Witness = {"number": 0, "scriptCode": ''}
                    # Pycharm IDE gives a type error for the following line, no idea why...
                    # noinspection PyTypeChecker
                    txobj["witness"].append(witness)
            script = mk_p2wpkh_scriptcode(pub)
            signing_tx = signature_form(txobj, i, script, self.hashcode, segwit=True)
            sig = ecdsa_tx_sign(signing_tx, priv, self.secondary_hashcode)
            if native_segwit:
                txobj["ins"][i]["script"] = ''
            else:
                txobj["ins"][i]["script"] = mk_p2wpkh_redeemscript(pub)
            witness: Witness = {"number": 2, "scriptCode": serialize_script([sig, pub])}
            # Pycharm IDE gives a type error for the following line, no idea why...
            # noinspection PyTypeChecker
            txobj["witness"].append(witness)
        else:
            if p2pk:
                script = mk_p2pk_script(pub)
            else:
                address = self.pubtoaddr(pub)
                script = addr_to_pubkey_script(address)
            signing_tx = signature_form(txobj, i, script, self.hashcode)
            sig = ecdsa_tx_sign(signing_tx, priv, self.hashcode)
            # Pycharm IDE gives a type error for the following line, no idea why...
            # noinspection PyTypeChecker
            script = serialize_script([sig]) if p2pk else serialize_script([sig, pub])
            txobj["ins"][i]["script"] = script
            if "witness" in txobj.keys():
                witness: Witness = {"number": 0, "scriptCode": ''}
                # Pycharm IDE gives a type error for the following line, no idea why...
                # noinspection PyTypeChecker
                txobj["witness"].append(witness)
        return txobj

    def signall(self, txobj: Union[str, Tx], priv: PrivateKeySignAllType) -> Tx:
        """
        Sign all inputs to a transaction using a private key.
        Priv is either a private key or a dictionary of address keys and private key values
        """
        if not isinstance(txobj, dict):
            txobj = deserialize(txobj)
        if isinstance(priv, dict):
            for i, inp in enumerate(txobj["ins"]):
                addr = inp['address']
                k = priv[addr]
                txobj = self.sign(txobj, i, k)
        else:
            for i in range(len(txobj["ins"])):
                txobj = self.sign(txobj, i, priv)
        return txobj

    def multisign(self, txobj: Union[str, Tx], i: int, script: str, priv) -> Tx:
        i = int(i)
        if not isinstance(txobj, dict):
            txobj = deserialize(txobj)
        if len(priv) <= 33:
            priv = safe_hexlify(priv)
        inp = txobj['ins'][i]
        segwit = False
        try:
            if address := inp['address']:
                segwit = self.is_native_segwit(address)
        except (IndexError, KeyError):
            pass
        return multisign(txobj, i, script, priv, self.hashcode, segwit=segwit)

    def mktx(self, ins: List[Union[TxInput, AnyStr]], outs: List[Union[TxOut, AnyStr]], locktime: int = 0,
             sequence: int = 0xFFFFFFFF) -> Tx:
        """[in0, in1...],[out0, out1...]

        Make an unsigned transaction from inputs and outputs. Change is not automatically included so any difference
        in value between inputs and outputs will be given as a miner's fee (transactions with too high fees will
        normally be blocked by Electrumx)

        Ins and outs are both lists of dicts.
        """

        txobj = {"locktime": locktime, "version": 1}
        for i, inp in enumerate(ins):
            if isinstance(inp, string_or_bytes_types):
                real_inp: TxInput = {"tx_hash": inp[:64], "tx_pos": int(inp[65:]), 'amount': 0}
                ins[i] = real_inp
                inp = real_inp
            if address := inp.get('address', ''):
                if self.segwit_supported and self.is_native_segwit(address):
                    txobj.update({"marker": 0, "flag": 1, "witness": []})
            inp.update({
                'script': '',
                'sequence': int(inp.get('sequence', sequence))
            })

        for i, out in enumerate(outs):
            if isinstance(out, string_or_bytes_types):
                o = out
                addr = o[:o.find(':')]
                val = int(o[o.find(':') + 1:])
                out = {}
                if is_hex(addr):
                    out["script"] = addr
                else:
                    out["address"] = addr
                out["value"] = val
                outs[i] = out
            if address := out.pop('address', None):
                out["script"] = self.addrtoscript(address)
            elif "script" not in out.keys():
                raise Exception("Could not find 'address' or 'script' in output.")
        txobj.update({'ins': ins, 'outs': outs})
        return txobj

    async def mktx_with_change(self, ins: List[Union[TxInput, AnyStr, ElectrumXTx]], outs: List[Union[TxOut, AnyStr]],
                               change_addr: str = None, fee: int = None, estimate_fee_blocks: int = 6,
                               locktime: int = 0, sequence: int = 0xFFFFFFFF) -> Tx:
        """[in0, in1...],[out0, out1...]

        Make an unsigned transaction from inputs, outputs change address and fee. A change output will be added with
        change sent to the change address..
        """
        change_addr = change_addr or ins[0]['address']
        isum = sum(inp['value'] for inp in ins)
        osum = sum(out['value'] for out in outs)
        change_out = {"address": change_addr, "value": isum - osum}
        outs += [change_out]
        txobj = self.mktx(ins, outs, locktime=locktime, sequence=sequence)

        if fee is None:
            fee = max(await self.estimate_fee(txobj, numblocks=estimate_fee_blocks), self.minimum_fee)
        if isum < osum + fee:
            raise Exception(f"Not enough money. You have {isum} but need {osum+fee} ({osum} + fee of {fee}).")

        if change_out['value'] > fee:
            change_out['value'] = isum - osum - fee
        else:
            outs.remove(change_out)

        return txobj

    async def preparemultitx(self, frm: str, outs: List[TxOut], change_addr: str = None,
                             fee: int = None, estimate_fee_blocks: int = 6) -> Tx:
        """
        Prepare transaction with multiple outputs, with change sent to from address
        """
        outvalue = int(sum(out['value'] for out in outs))
        unspents = await self.unspent(frm)
        if fee is None:
            unspents2 = select(unspents, outvalue)
            tx = await self.mktx_with_change(unspents2, deepcopy(outs), fee=0, change_addr=change_addr)
            fee = max(await self.estimate_fee(tx, estimate_fee_blocks), self.minimum_fee)
        unspents2 = select(unspents, outvalue + fee)
        change_addr = change_addr or frm
        return await self.mktx_with_change(unspents2, outs, fee=fee, change_addr=change_addr,
                                           estimate_fee_blocks=estimate_fee_blocks)

    async def preparetx(self, frm: str, to: str, value: int, fee: int = None, estimate_fee_blocks: int = 6,
                        change_addr: str = None) -> Tx:
        """
        Prepare a transaction using from and to addresses, value and a fee, with change sent back to from address
        """
        outs: List[TxOut] = [{'address': to, 'value': value}]
        return await self.preparemultitx(frm, outs, fee=fee, estimate_fee_blocks=estimate_fee_blocks,
                                         change_addr=change_addr)

    async def preparesignedmultirecipienttx(self, privkey: PrivateKeySignAllType, frm: str, outs: List[TxOut],
                                            change_addr: str = None, fee: int = None, estimate_fee_blocks: int = 6) -> Tx:
        """
        Prepare transaction with multiple outputs, with change sent back to from address or given change_addr
        Requires private key, address:value pairs and optionally the change address and fee
        segwit paramater specifies that the inputs belong to a segwit address
        addr, if provided, will explicity set the from address, overriding the auto-detection of the address from the
        private key.It will also be used, along with the privkey to automatically detect a segwit transaction for coins
        which support segwit, overriding the segwit kw
        """
        tx = await self.preparemultitx(frm, outs, fee=fee, change_addr=change_addr,
                                       estimate_fee_blocks=estimate_fee_blocks)
        tx2 = self.signall(tx, privkey)
        return tx2

    async def preparesignedtx(self, privkey: PrivateKeySignAllType, frm: str, to: str, value: int,
                              change_addr: str = None, fee: int = None, estimate_fee_blocks: int = 6) -> Tx:
        """
        Prepare a tx with a specific amount from address belonging to private key to another address, returning change
        to the from address or change address, if set.
        Requires private key, target address, value and optionally the change address and fee
        segwit paramater specifies that the inputs belong to a segwit address
        addr, if provided, will explicity set the from address, overriding the auto-detection of the address from the
        private key.It will also be used, along with the privkey, to automatically detect a segwit transaction for
        coins which support segwit, overriding the segwit kw
        """
        outs = [{'address': to, 'value': value}]
        return await self.preparesignedmultirecipienttx(privkey, frm, outs, change_addr=change_addr,
                                                        fee=fee, estimate_fee_blocks=estimate_fee_blocks)

    async def send_to_multiple_receivers_tx(self, privkey: PrivateKeySignAllType, addr: str, outs: List[TxOut],
                                            change_addr: str = None, fee: int = None, estimate_fee_blocks: int = 6):
        """
        Send transaction with multiple outputs, with change sent back to from addrss
        Requires private key, address:value pairs and optionally the change address and fee
        segwit paramater specifies that the inputs belong to a segwit address
        addr, if provided, will explicity set the from address, overriding the auto-detection of the address from the
        private key.It will also be used, along with the privkey to automatically detect a segwit transaction for
        coins which support segwit, overriding the segwit kw
        """
        tx = await self.preparesignedmultirecipienttx(privkey, addr, outs, change_addr=change_addr,
                                                      fee=fee, estimate_fee_blocks=estimate_fee_blocks)
        return await self.pushtx(tx)

    async def send(self, privkey: PrivateKeySignAllType, frm: str, to: str, value: int, change_addr: str = None,
                   fee: int = None, estimate_fee_blocks: int = 6):
        """
        Send a specific amount from address belonging to private key to another address, returning change to the
        from address or change address, if set.
        Requires private key, target address, value and optionally the change address and fee
        segwit paramater specifies that the inputs belong to a segwit address
        addr, if provided, will explicity set the from address, overriding the auto-detection of the address from the
        private key.It will also be used, along with the privkey, to automatically detect a segwit transaction for
        coins which support segwit, overriding the segwit kw
        """
        outs = [{'address': to, 'value': value}]
        return await self.send_to_multiple_receivers_tx(privkey, frm, outs, change_addr=change_addr,
                                                        fee=fee, estimate_fee_blocks=estimate_fee_blocks)

    async def inspect(self, tx: Union[str, Tx]) -> TXInspectType:
        if not isinstance(tx, dict):
            tx = deserialize(tx)
        isum = 0
        ins = {}
        for _in in tx['ins']:
            h = _in['tx_hash']
            i = _in['tx_pos']
            prevout = (await anext(self.get_txs(h)))['outs'][i]
            isum += prevout['value']
            a = self.scripttoaddr(prevout['script'])
            ins[a] = ins.get(a, 0) + prevout['value']
        outs = []
        osum = 0
        for _out in tx['outs']:
            outs.append({'address': self.scripttoaddr(_out['script']),
                         'value': _out['value']})
            osum += _out['value']
        return {
            'fee': isum - osum,
            'outs': outs,
            'ins': ins
        }

    async def wait_unspents_changed(self, addr: str, start_unspents: ElectrumXUnspentResponse):
        unspents = start_unspents
        while unspents == start_unspents:
            unspents = await self.unspent(addr)
            if start_unspents == unspents:
                await asyncio.sleep(1)

    def wallet(self, seed: str, passphrase: str = None, **kwargs) -> HDWallet:
        if not bip39_is_checksum_valid(seed) == (True, True):
            raise Exception("BIP39 Checksum failed. This is not a valid BIP39 seed")
        ks = standard_from_bip39_seed(seed, passphrase, coin=self)
        return HDWallet(ks, **kwargs)

    def watch_wallet(self, xpub, **kwargs) -> HDWallet:
        ks = from_xpub(xpub, self, 'p2pkh')
        return HDWallet(ks, **kwargs)

    def p2wpkh_p2sh_wallet(self, seed: str, passphrase: str = None, **kwargs) -> HDWallet:
        if not self.segwit_supported:
            raise Exception("P2WPKH-P2SH segwit not enabled for this coin")
        if not bip39_is_checksum_valid(seed) == (True, True):
            raise Exception("BIP39 Checksum failed. This is not a valid BIP39 seed")
        ks = p2wpkh_p2sh_from_bip39_seed(seed, passphrase, coin=self)
        return HDWallet(ks, **kwargs)

    def watch_p2wpkh_p2sh_wallet(self, xpub, **kwargs) -> HDWallet:
        ks = from_xpub(xpub, self, 'p2wpkh-p2sh')
        return HDWallet(ks, **kwargs)

    def p2wpkh_wallet(self, seed: str, passphrase: str = None, **kwargs) -> HDWallet:
        if not bip39_is_checksum_valid(seed) == (True, True):
            raise Exception("BIP39 Checksum failed. This is not a valid BIP39 seed")
        ks = p2wpkh_from_bip39_seed(seed, passphrase, coin=self)
        return HDWallet(ks, **kwargs)

    def watch_p2wpkh_wallet(self, xpub: str, **kwargs) -> HDWallet:
        ks = from_xpub(xpub, self, 'p2wpkh')
        return HDWallet(ks, **kwargs)

    def electrum_wallet(self, seed: str, passphrase: str = None, **kwargs) -> HDWallet:
        ks = from_electrum_seed(seed, passphrase, False, coin=self)
        return HDWallet(ks, **kwargs)

    def watch_electrum_wallet(self, xpub: str, **kwargs) -> HDWallet:
        ks = from_xpub(xpub, self, 'p2pkh', electrum=True)
        return HDWallet(ks, **kwargs)

    def watch_electrum_p2wpkh_wallet(self, xpub: str, **kwargs) -> HDWallet:
        ks = from_xpub(xpub, self, 'p2wpkh', electrum=True)
        return HDWallet(ks, **kwargs)



import asyncio
import os.path

from pathlib import Path
from packaging.version import parse as parse_version

import certifi
import ssl
import itertools
import aiorpcx
from .. import constants
from aiorpcx import RPCSession, Notification, NewlineFramer
from aiorpcx.curio import TaskTimeout
from aiorpcx.rawsocket import RSClient
from collections import defaultdict
import json
import random
from typing import List, Dict, Any, Optional, Callable
from .types import (ElectrumXBlockResponse, ElectrumXBlockHeadersResponse, BlockHeaderNotificationCallback,
    ElectrumXBalanceResponse, ElectrumXHistoryResponse, ElectrumXMempoolResponse, ElectrumXUnspentResponse,
    AddressNotificationCallback, ElectrumXGetTxResponse, ElectrumXMerkleResponse)


ca_path = certifi.where()

MAX_INCOMING_MSG_SIZE = 1_000_000  # in bytes
_KNOWN_NETWORK_PROTOCOLS = {'t', 's'}
PREFERRED_NETWORK_PROTOCOL = 's'
assert PREFERRED_NETWORK_PROTOCOL in _KNOWN_NETWORK_PROTOCOLS


class NetworkException(Exception):
    pass


class RequestCorrupted(Exception):
    pass


class ErrorParsingSSLCert(Exception):
    pass


class ErrorGettingSSLCertFromServer(Exception):
    pass


class ErrorSSLCertFingerprintMismatch(Exception):
    pass


class InvalidOptionCombination(Exception):
    pass


class ConnectError(NetworkException):
    pass


class ProtocolNotSupportedError(BaseException):
    pass


class _RSClient(RSClient):
    async def create_connection(self):
        try:
            return await asyncio.wait_for(super().create_connection(), timeout=60)
        except OSError as e:
            # note: using "from e" here will set __cause__ of ConnectError
            raise ConnectError(e) from e


def read_json(filename, default):
    path = os.path.join(os.path.dirname(__file__), filename)
    try:
        with open(path, 'r') as f:
            r = json.loads(f.read())
    except:
        r = default
    return r


class NetworkException(Exception):
    pass


class GracefulDisconnect(NetworkException):
    pass


class RequestTimedOut(GracefulDisconnect):
    def __str__(self):
        return "Network request timed out."


class NetworkTimeout:
    # seconds
    class Generic:
        NORMAL = 30
        RELAXED = 45
        MOST_RELAXED = 600

    class Urgent(Generic):
        NORMAL = 10
        RELAXED = 20
        MOST_RELAXED = 60


class NotificationSession(RPCSession):

    def __init__(self, *args, **kwargs):
        super(NotificationSession, self).__init__(*args, **kwargs)
        self.subscriptions = defaultdict(list)
        self.cache = {}
        self.default_timeout = NetworkTimeout.Generic.NORMAL
        self._msg_counter = itertools.count(start=1)
        self.cost_hard_limit = 0  # disable aiorpcx resource limits

    async def handle_request(self, request):
        try:
            if isinstance(request, Notification):
                params, result = request.args[:-1], request.args[-1]
                key = self.get_hashable_key_for_rpc_call(request.method, params)
                if key in self.subscriptions:
                    self.cache[key] = result
                    for queue in self.subscriptions[key]:
                        await queue.put(request.args)
                else:
                    raise Exception(f'unexpected notification')
            else:
                raise Exception(f'unexpected request. not a notification')
        except Exception as e:
            await self.close()
            raise

    async def send_request(self, *args, timeout: int = None, **kwargs):
        # note: semaphores/timeouts/backpressure etc are handled by
        # aiorpcx. the timeout arg here in most cases should not be set
        msg_id = next(self._msg_counter)
        try:
            # note: RPCSession.send_request raises TaskTimeout in case of a timeout.
            # TaskTimeout is a subclass of CancelledError, which is *suppressed* in TaskGroups
            response = await asyncio.wait_for(
                super().send_request(*args, **kwargs),
                timeout)
        except (TaskTimeout, asyncio.TimeoutError) as e:
            raise RequestTimedOut(f'request timed out: {args} (id: {msg_id})') from e
        else:
            return response

    def set_default_timeout(self, timeout):
        self.sent_request_timeout = timeout
        self.max_send_delay = timeout

    async def subscribe(self, method: str, params: List, queue: asyncio.Queue):
        # note: until the cache is written for the first time,
        # each 'subscribe' call might make a request on the network.
        key = self.get_hashable_key_for_rpc_call(method, params)
        self.subscriptions[key].append(queue)
        if key in self.cache:
            result = self.cache[key]
        else:
            result = await self.send_request(method, params)
            self.cache[key] = result
        await queue.put(params + [result])

    def unsubscribe(self, queue):
        """Unsubscribe a callback to free object references to enable GC."""
        # note: we can't unsubscribe from the server, so we keep receiving
        # subsequent notifications
        for v in self.subscriptions.values():
            if queue in v:
                v.remove(queue)

    @classmethod
    def get_hashable_key_for_rpc_call(cls, method, params):
        """Hashable index for subscriptions and cache"""
        return str(method) + repr(params)

    def default_framer(self):
        return NewlineFramer(max_size=MAX_INCOMING_MSG_SIZE)

    async def close(self, *, force_after: int = None):
        """Closes the connection and waits for it to be closed.
        We try to flush buffered data to the wire, which can take some time.
        """
        if force_after is None:
            # We give up after a while and just abort the connection.
            # Note: specifically if the server is running Fulcrum, waiting seems hopeless,
            #       the connection must be aborted (see https://github.com/cculianu/Fulcrum/issues/76)
            # Note: if the ethernet cable was pulled or wifi disconnected, that too might
            #       wait until this timeout is triggered
            force_after = 1  # seconds
        await super().close(force_after=force_after)


class CannotConnectToAnyElectrumXServer(BaseException):
    def __init__(self):
        self.message = "Unable to connect to any ElectrumX Server"
        super().__init__(self.message)


class ElectrumXClient:
    host: str = None
    port: int = None
    server: Dict[str, Any]
    session: Optional[NotificationSession] = None
    requires_scripthash: bool = True

    def __init__(self, server_file: str = "bitcoin.json", connection_timeout: int = 5,
                 use_ssl: bool = True, tor: bool = False, client_name: str = constants.CLIENT_NAME,
                 ping_interval: int = 30, accept_self_signed_certs: bool = True):
        self._active_subscriptions: Dict[str, List[asyncio.Task]] = {}
        self._tasks = []
        self.restart_condition = asyncio.Condition()
        self._connection_task: Optional[asyncio.Task] = None
        self._use_ssl = use_ssl
        self._tor = tor
        self._accept_self_signed_certs = accept_self_signed_certs
        self.cert_path: Optional[Path] = None
        self.connection_timeout = connection_timeout
        servers = read_json(f"servers/{server_file}", {})
        self._port_key = "s" if self._use_ssl else "t"
        self._servers = {k: v for k, v in servers.items() if v.get(self._port_key)}
        self._lock = asyncio.Lock()
        self.is_closing = False
        self._failed_servers: List[str] = []
        self.version = constants.PROTOCOL_VERSION
        self.server_version: Optional[List[str]] = None
        self.client_name = client_name
        self._ping_interval = ping_interval

    def compare_versions(self, min_version: str) -> bool:
        if self.server_version:
            version = self.server_version[1]
            server = parse_version(version)
            minimum = parse_version(min_version)
            return server >= minimum
        return False

    def _get_eligible_servers(self) -> Dict[str, Any]:
        return {k: v for k, v in self._servers.items() if k not in self._failed_servers and self._port_key in v.keys() and
                (self._tor and k.endswith('onion') or (not self._tor and not k.endswith('onion')))}

    def _choose_new_server(self) -> str:
        eligible = self._get_eligible_servers()
        return random.choice(list(eligible.keys())) if eligible else None

    def _set_new_server(self):
        self.host = self._choose_new_server()
        self.server = self._servers[self.host]
        try:
            self.port = int(self.server[self._port_key])
        except KeyError:
            raise ProtocolNotSupportedError

    async def _get_ssl_context(self) -> Optional[ssl.SSLContext]:
        if not self._use_ssl:
            # using plaintext TCP
            return None

        # see if we already have cert for this server; or get it for the first time
        ca_sslc = ssl.create_default_context(purpose=ssl.Purpose.SERVER_AUTH, cafile=ca_path)
        if self._accept_self_signed_certs:
            ca_sslc.check_hostname = False
            ca_sslc.verify_mode = ssl.CERT_NONE
        return ca_sslc

    async def monitor_connection(self):
        i = 0
        while True:
            await asyncio.sleep(1)
            i += 1
            # If the session/transport is no longer open, we disconnect.
            # e.g. if the remote cleanly sends EOF, we would handle that here.
            # note: If the user pulls the ethernet cable or disconnects wifi,
            #       ideally we would detect that here, so that the GUI/etc can reflect that.
            #       - On Android, this seems to work reliably , where asyncio.BaseProtocol.connection_lost()
            #         gets called with e.g. ConnectionAbortedError(103, 'Software caused connection abort').
            #       - On desktop Linux/Win, it seems BaseProtocol.connection_lost() is not called in such cases.
            #         Hence, in practice the connection issue will only be detected the next time we try
            #         to send a message (plus timeout), which can take minutes...
            if not self.session or self.session.is_closing():
                raise GracefulDisconnect('session was closed')
            elif i == self._ping_interval:
                try:
                    await self._send_request("server.ping", timeout=self.connection_timeout)
                    i = 0
                except TimeoutError:
                    raise GracefulDisconnect('session was closed')

    async def _open_session(self, sslc: Optional[ssl.SSLContext] = None) -> None:
        session_factory = lambda *args, **kwargs: NotificationSession(*args, **kwargs)
        async with _RSClient(session_factory=session_factory,
                             host=self.host, port=self.port,
                             ssl=sslc) as session:
            self.session = session

            self.session.set_default_timeout(NetworkTimeout.Generic.NORMAL)

            self.server_version = await self._send_request("server.version", self.client_name, self.version, timeout=10)
            async with self.restart_condition:
                self.restart_condition.notify_all()
            await self.monitor_connection()
        self.session = None

    async def _connect(self) -> None:
        sslc = await self._get_ssl_context()
        await self._open_session(sslc=sslc)

    async def redo_connection(self):
        await self.session.close()
        await self.wait_new_start()

    async def _on_connection_failure(self):
        self.session = None
        self._failed_servers.append(self.host)
        if not self.is_closing:
            if len(self._failed_servers) == len(self._servers):
                raise CannotConnectToAnyElectrumXServer
            await self.connect_to_any_server()

    async def connect_to_any_server(self) -> None:
        self._set_new_server()
        connect_task = asyncio.create_task(self._connect())

        try:
            await asyncio.wait_for(self.wait_new_start(), timeout=5)
        except asyncio.TimeoutError:
            if not connect_task.done():
                connect_task.cancel()
        try:
            await connect_task
        except (asyncio.TimeoutError, asyncio.CancelledError, aiorpcx.jsonrpc.RPCError, OSError,
                GracefulDisconnect, ConnectError, ProtocolNotSupportedError, ssl.SSLError) as e:
            await self._on_connection_failure()

    async def wait_new_start(self):
        async with self.restart_condition:
            await self.restart_condition.wait()

    async def _ensure_connected(self):
        async with self._lock:
            if not self.is_closing:
                if not self.session:
                    """
                    Wait until successful connection or connection task completes without any successful connections
                    """
                    self._connection_task = asyncio.create_task(self.connect_to_any_server(), name="connection_task")
                    is_connected_task = asyncio.create_task(self.wait_new_start(), name="is_connected_task")
                    done, pending = await asyncio.wait([is_connected_task, self._connection_task],
                                                       return_when=asyncio.FIRST_COMPLETED)

                    for task in done:
                        if not task.cancelled():
                            if exc := task.exception():
                                for p in pending:
                                    p.cancel()
                                raise exc
            else:
                raise ConnectionError("JSONRPC Connection is already closing, cannot send message")

    async def cancel_subscriptions(self) -> None:
        async with self._lock:
            subscription_tasks = list(itertools.chain(*self._active_subscriptions.values()))
            for task in subscription_tasks:
                task.cancel()
            all_tasks = subscription_tasks + self._tasks
            if all_tasks:
                await asyncio.wait(all_tasks)

    async def close(self):
        self.is_closing = True
        await self.cancel_subscriptions()
        if self.session:
            await self.session.close(force_after=5)
        if self._connection_task:
            try:
                await asyncio.wait_for(self._connection_task, timeout=self.connection_timeout)
            except (TimeoutError, OSError, ConnectError) as e:
                if not self._connection_task.done():
                    self._connection_task.cancel()
            await self._connection_task

    async def _send_request(self, method: str, *args, timeout: int = 30, **kwargs) -> Any:
        return await self.session.send_request(method, args, timeout=timeout, **kwargs)

    async def send_request(self, method: str, *args, timeout: int = 30, **kwargs) -> Any:
        await self._ensure_connected()
        return await self._send_request(method, *args, timeout=timeout, **kwargs)

    def _on_task_complete(self, task: asyncio.Task) -> None:
        if task in self._tasks:
            self._tasks.remove(task)
        if not task.cancelled():
            if exc := task.exception():
                raise exc

    def _on_subscription_task_complete(self, task: asyncio.Task) -> None:
        remove_method = None
        try:
            for method, tasks in self._active_subscriptions.items():
                if task in tasks:
                    tasks.remove(task)
                    if not tasks:
                        remove_method = method
                    break
            if not task.cancelled():
                if exc := task.exception():
                    raise exc
        finally:
            if remove_method:
                del self._active_subscriptions[remove_method]

    async def _subscribe(self, method: str, callback: Callable, *args) -> None:
        queue = asyncio.Queue()
        session = None
        subscribed = False
        just_restarted = False
        last_item = None
        wait_next_start_task = None
        try:
            while True:
                if not subscribed:
                    session = self.session
                    await session.subscribe(method, list(args), queue)
                    subscribed = True
                    just_restarted = True
                    wait_next_start_task = asyncio.create_task(self.wait_new_start())
                queue_task = asyncio.create_task(queue.get())
                done, pending = await asyncio.wait([queue_task, wait_next_start_task],
                                                   return_when=asyncio.FIRST_COMPLETED)
                if wait_next_start_task in done:
                    subscribed = False
                if queue_task in done:
                    item = queue_task.result()
                    """
                    After restarting a subscription we want to avoid processing the item we receive immediately,
                    which was most likely already received on the last connection
                    """
                    if not just_restarted or item != last_item:
                        last_item = item
                        task = asyncio.create_task(callback(*item))
                        self._tasks.append(task)
                        task.add_done_callback(self._on_task_complete)
                    just_restarted = False
                else:
                    queue_task.cancel()
        finally:
            if session:
                session.unsubscribe(queue)
                if "scripthash" in method and not self.is_closing and self.compare_versions("1.4.2"):
                    await self.send_request("blockchain.scripthash.unsubscribe", *args)

    @staticmethod
    def _get_sub_name(method, *args):
        arg0 = " ".join(args)
        return f'{method}[{arg0}]'

    async def unsubscribe(self, method: str, *args):
        name = self._get_sub_name(method, *args)
        tasks = self._active_subscriptions[name]
        if tasks:
            for task in tasks:
                task.cancel()
        await asyncio.wait(tasks)

    def _create_subscribe_task(self, method: str, callback: Callable, *args) -> None:
        task = asyncio.create_task(self._subscribe(method, callback, *args))
        name = self._get_sub_name(method, *args)
        if not self._active_subscriptions.get(name):
            self._active_subscriptions[name] = []
        self._active_subscriptions[name].append(task)
        task.add_done_callback(self._on_subscription_task_complete)

    async def subscribe(self, callback: Callable, method: str, *args) -> None:
        await self._ensure_connected()
        self._create_subscribe_task(method, callback, *args)

    async def block_header(self, height: int, cp_height: int = 0) -> ElectrumXBlockResponse:
        return await self.send_request("blockchain.block.header", height, cp_height)

    async def block_headers(self, start_height: int, count: int, cp_height: int = 0) -> ElectrumXBlockHeadersResponse:
        return await self.send_request("blockchain.block.headers", start_height, count, cp_height)

    async def estimate_fee(self, numblocks: int = 6) -> float:
        return await self.send_request("blockchain.estimatefee", numblocks)

    async def relay_fee(self) -> float:
        return await self.send_request("blockchain.relayfee")

    async def subscribe_to_block_headers(self, callback: BlockHeaderNotificationCallback) -> None:
        await self.subscribe(callback, "blockchain.headers.subscribe")

    async def unsubscribe_from_block_headers(self) -> None:
        await self.unsubscribe("blockchain.headers.subscribe")

    async def get_balance(self, scripthash: str) -> ElectrumXBalanceResponse:
        return await self.send_request("blockchain.scripthash.get_balance", scripthash)

    async def get_history(self, scripthash: str) -> ElectrumXHistoryResponse:
        return await self.send_request("blockchain.scripthash.get_history", scripthash)

    async def get_mempool(self, scripthash: str) -> ElectrumXMempoolResponse:
        return await self.send_request("blockchain.scripthash.get_mempool", scripthash)

    async def unspent(self, scripthash: str) -> ElectrumXUnspentResponse:
        return await self.send_request("blockchain.scripthash.listunspent", scripthash)

    async def subscribe_to_address(self, callback: AddressNotificationCallback, scripthash: str) -> None:
        await self.subscribe(callback, "blockchain.scripthash.subscribe", scripthash)

    async def unsubscribe_from_address(self, scripthash: str) -> None:
        await self.unsubscribe("blockchain.scripthash.subscribe", scripthash)

    async def broadcast_tx(self, raw_tx: str) -> str:
        return await self.send_request("blockchain.transaction.broadcast", raw_tx)

    async def get_tx(self, tx_hash: str, verbose: bool = False) -> ElectrumXGetTxResponse:
        try:
            return await self.send_request("blockchain.transaction.get", tx_hash, verbose)
        except aiorpcx.jsonrpc.ProtocolError as e:
            if any(msg in e.message for msg in ("verbose transactions are currently unsupported",)):
                "Some servers return this even if later than v 1.2 when verbose transactions were introduced"
                await self.redo_connection()
                return await self.get_tx(tx_hash, verbose=verbose)
            raise e

    async def get_merkle(self, tx_hash: str, height: int) -> Optional[ElectrumXMerkleResponse]:
        if height <= 0:
            return None     # Transaction not in blockchain yet
        try:
            return await self.send_request("blockchain.transaction.get_merkle", tx_hash, height)
        except aiorpcx.jsonrpc.RPCError as e:
            if any(msg in e.message for msg in ("No confirmed transaction", "unconfirmed")):
                return None
            raise e

    async def get_donation_address(self) -> str:
        return await self.send_request("server.donation_address")



import asyncio
from concurrent.futures import Future
import threading
import inspect
import janus
from ..coins_async.base import BaseCoin
from cryptos.utils import alist
from ..wallet import HDWallet
from ..electrumx_client.types import (ElectrumXBalanceResponse, ElectrumXMultiBalanceResponse, ElectrumXTx,
                                      ElectrumXMerkleResponse, ElectrumXUnspentResponse, ElectrumXMultiTxResponse,
                                      ElectrumXHistoryResponse)
from ..types import (Tx, BlockHeader, BlockHeaderCallbackSync, AddressCallbackSync, AddressTXCallbackSync, MerkleProof,
                     AddressBalance, TxOut, TxInput, TXInspectType, PrivateKeySignAllType, PrivkeyType, PubKeyType)
from typing import Optional, Tuple, Any, List, Union, Dict, AnyStr, Type


class BaseSyncCoin:
    timeout: int = 10
    is_closing: bool = False
    coin_class: Type[BaseCoin]
    _thread: threading.Thread = None

    """
    This class wraps the parent class async coroutine methods in synchronous methods.
    A new thread is created to run the parent class async coroutine methods in an asyncio event loop.
    The thread-safe janus.Queue is used to pass requests to the thread running the event loop.
    concurrent.futures.Future is used to return results and exceptions to the main thread.
    
    For comments explaining each method, check the async coin class.
    """

    def __init__(self, testnet: bool = False, use_ssl: bool = None, **kwargs):
        self._async_coin = self.coin_class(testnet=testnet, use_ssl=use_ssl, **kwargs)
        self._request_queue: Optional[janus.Queue[Tuple[Future, str, tuple, dict[str, Any]]]] = None
        self._loop_is_started = threading.Event()

    def __getattr__(self, item):
        return getattr(self._async_coin, item)

    def start(self):
        if not self._thread or not self._thread.is_alive():
            self._thread = threading.Thread(target=self.start_event_loop, daemon=True)
            self._thread.start()
        self._loop_is_started.wait(timeout=10)

    def start_event_loop(self):
        asyncio.run(self.run())

    async def run(self):
        self._request_queue = janus.Queue()
        fut: Optional[Future]
        method: str
        args: tuple
        kwargs: dict
        if not self.is_closing:
            try:
                asyncio.get_running_loop().call_soon(self._loop_is_started.set)
                while True:
                    val = await self._request_queue.async_q.get()
                    fut, method, args, kwargs = val
                    try:
                        if method == "_property":
                            coro = getattr(self._async_coin, args[0])
                        else:
                            coro = getattr(self._async_coin, method)(*args, **kwargs)
                            if inspect.isasyncgen(coro):
                                coro = alist(coro)
                        result = await asyncio.wait_for(coro, timeout=self.timeout)
                        fut.set_result(result)
                    except BaseException as e:
                        fut.set_exception(e)
                    finally:
                        if method == "close":
                            break
            finally:
                if not self._async_coin.is_closing:
                    await self._async_coin.close()
            self._loop_is_started.clear()

    def _run_async(self, method: str, *args, **kwargs):
        self.start()
        fut = Future()
        self._request_queue.sync_q.put((fut, method, args, kwargs))
        return fut.result(timeout=self.timeout * 2)

    def __del__(self):
        self.close()

    def close(self):
        self.is_closing = True
        if self._loop_is_started.is_set():
            fut = Future()
            self._request_queue.sync_q.put((fut, "close", (), {}))
            fut.result(timeout=100)

    def tx_size(self, txobj: Tx) -> float:
        return self._async_coin.tx_size(txobj)

    def estimate_fee_per_kb(self, numblocks: int = 6) -> float:
        return self._run_async("estimate_fee_per_kb", numblocks=numblocks)

    def estimate_fee(self, txobj: Tx, numblocks: int = 6) -> int:
        return self._run_async("estimate_fee", txobj, numblocks=numblocks)

    def raw_block_header(self, height: int) -> str:
        return self._run_async("raw_block_header", height)

    def block_header(self, height: int) -> BlockHeader:
        return self._run_async("block_header", height)

    def block_headers(self, *args: int) -> List[BlockHeader]:
        return self._run_async("block_headers", *args)

    def subscribe_to_block_headers(self, callback: BlockHeaderCallbackSync) -> None:
        return self._run_async("subscribe_to_block_headers", callback)

    def unsubscribe_from_block_headers(self) -> None:
        return self._run_async("unsubscribe_from_block_headers")

    @property
    def block(self) -> Tuple[int, str, BlockHeader]:
        return self._run_async("_property", "block")

    def confirmations(self, height: int) -> int:
        return self._run_async("confirmations", height)

    def subscribe_to_address(self,  callback: AddressCallbackSync, addr: str) -> None:
        return self._run_async("subscribe_to_address", callback, addr)

    def unsubscribe_from_address(self, addr: str) -> None:
        return self._run_async("unsubscribe_from_address", addr)

    def subscribe_to_address_transactions(self, callback: AddressTXCallbackSync, addr: str) -> None:
        return self._run_async("subscribe_to_address_transactions", callback, addr)

    def get_balance(self, addr: str) -> ElectrumXBalanceResponse:
        return self._run_async("get_balance", addr)

    def get_balances(self, *args: str) -> List[ElectrumXMultiBalanceResponse]:
        return self._run_async("get_balances", *args)

    def get_merkle(self, tx: ElectrumXTx) -> Optional[ElectrumXMerkleResponse]:
        return self._run_async("get_merkle", tx)

    def merkle_prove(self, tx: ElectrumXTx) -> MerkleProof:
        return self._run_async("merkle_prove", tx)

    def merkle_prove_by_txid(self, tx_hash: str) -> MerkleProof:
        return self._run_async("merkle_prove_by_txid", tx_hash)

    def unspent(self, addr: str, merkle_proof: bool = False) -> ElectrumXUnspentResponse:
        return self._run_async("unspent", addr, merkle_proof=merkle_proof)

    def get_unspents(self, *args: str, merkle_proof: bool = False) -> List[ElectrumXMultiTxResponse]:
        return self._run_async("get_unspents", *args, merkle_proof=merkle_proof)

    def balance_merkle_proven(self, addr: str) -> int:
        return self._run_async("balance_merkle_proven", addr)

    def balances_merkle_proven(self, *args: str) -> List[AddressBalance]:
        return self._run_async("balances_merkle_proven", *args)

    def history(self, addr: str, merkle_proof: bool = False) -> ElectrumXHistoryResponse:
        return self._run_async("history", addr, merkle_proof=merkle_proof)

    def get_histories(self, *args: str, merkle_proof: bool = False) -> List[ElectrumXMultiTxResponse]:
        return self._run_async("get_histories", *args, merkle_proof=merkle_proof)

    def get_raw_tx(self, tx_hash: str) -> str:
        return self._run_async("get_raw_tx", tx_hash)

    def get_tx(self, tx_hash: str) -> Tx:
        return self._run_async("get_tx", tx_hash)

    def get_verbose_tx(self, tx_hash: str) -> Dict[str, Any]:       # Make TypedDict
        return self._run_async("get_verbose_tx", tx_hash)

    def get_txs(self, *args: str) -> List[Tx]:
        return self._run_async("get_txs", *args)

    def pushtx(self, tx: Union[str, Tx]):
        return self._run_async("pushtx", tx)

    def privtopub(self, privkey: PrivkeyType) -> str:
        return self._async_coin.privtopub(privkey)

    def pubtoaddr(self, pubkey: PubKeyType) -> str:
        return self._async_coin.pubtoaddr(pubkey)

    def privtoaddr(self, privkey: PrivkeyType) -> str:
        return self._async_coin.privtoaddr(privkey)

    def electrum_address(self, masterkey: AnyStr, n: int, for_change: int = 0) -> str:
        return self._async_coin.electrum_address(masterkey, n, for_change=for_change)

    def encode_privkey(self, privkey: PrivkeyType, formt: str, script_type: str = "p2pkh") -> PrivkeyType:
        return self._async_coin.encode_privkey(privkey, formt, script_type)

    def is_p2pkh(self, addr: str) -> bool:
        return self._async_coin.is_p2pkh(addr)

    def is_p2sh(self, addr: str) -> bool:
        return self._async_coin.is_p2sh(addr)

    def is_native_segwit(self, addr: str) -> bool:
        return self._async_coin.is_native_segwit(addr)

    def is_address(self, addr: str) -> bool:
        return self._async_coin.is_address(addr)

    def is_legacy_segwit_or_multisig(self, addr: str) -> bool:
        return self._async_coin.maybe_legacy_segwit(addr)

    def is_segwit_or_multisig(self, addr: str) -> bool:
        return self._async_coin.is_segwit_or_p2sh(addr)

    def output_script_to_address(self, script: str) -> str:
        return self._async_coin.output_script_to_address(script)

    def scripttoaddr(self, script: str) -> str:
        return self._async_coin.scripttoaddr(script)

    def p2sh_scriptaddr(self, script: str) -> str:
        return self._async_coin.p2sh_scriptaddr(script)

    def addrtoscript(self, addr: str) -> str:
        return self._async_coin.addrtoscript(addr)

    def addrtoscripthash(self, addr: str) -> str:
        return self._async_coin.addrtoscripthash(addr)

    def pubtop2w(self, pub: str) -> str:
        return self._async_coin.pubtop2wpkh_p2sh(pub)

    def hash_to_segwit_addr(self, pub_hash: str) -> str:
        return self._async_coin.hash_to_segwit_addr(pub_hash)

    def pub_to_segwit_address(self, pubkey) -> str:
        return self._async_coin.pub_to_segwit_address(pubkey)

    def script_to_p2wsh(self, script) -> str:
        return self._async_coin.script_to_p2wsh(script)

    def mk_multsig_address(self, *args: str, num_required: int = None) -> Tuple[str, str]:
        return self._async_coin.mk_multsig_address(*args, num_required=num_required)

    def sign(self, txobj: Union[Tx, AnyStr], i: int, priv: PrivkeyType) -> Tx:
        return self._async_coin.sign(txobj, i, priv)

    def signall(self, txobj: Union[str, Tx], priv: PrivateKeySignAllType) -> Tx:
        return self._async_coin.signall(txobj, priv)

    def multisign(self, tx: Union[str, Tx], i: int, script: str, pk) -> Tx:
        return self._async_coin.multisign(tx, i, script, pk)

    def mktx(self, ins: List[Union[TxInput, AnyStr]], outs: List[Union[TxOut, AnyStr]], locktime: int = 0,
             sequence: int = 0xFFFFFFFF) -> Tx:
        return self._async_coin.mktx(ins, outs, locktime=locktime, sequence=sequence)

    def mktx(self, ins: List[Union[TxInput, AnyStr]], outs: List[Union[TxOut, AnyStr]], locktime: int = 0,
             sequence: int = 0xFFFFFFFF) -> Tx:
        return self._async_coin.mktx(ins, outs, locktime=locktime, sequence=sequence)

    def mktx_with_change(self, ins: List[Union[TxInput, AnyStr, ElectrumXTx]], outs: List[Union[TxOut, AnyStr]],
                         change_addr: str = None, fee: int = None, estimate_fee_blocks: int = 6, locktime: int = 0,
                         sequence: int = 0xFFFFFFFF) -> Tx:
        return self._run_async("mktx_with_change", ins, outs, change_addr=change_addr, fee=fee,
                               estimate_fee_blocks=estimate_fee_blocks, locktime=locktime, sequence=sequence)

    def preparemultitx(self, frm: str, outs: List[TxOut], change_addr: str = None,
                       fee: int = None, estimate_fee_blocks: int = 6) -> Tx:
        return self._run_async("preparemultitx", frm, outs, change_addr=change, fee=fee,
                               estimate_fee_blocks=estimate_fee_blocks)

    def preparetx(self, frm: str, to: str, value: int, fee: int = None, estimate_fee_blocks: int = 6,
                  change_addr: str = None) -> Tx:
        return self._run_async("preparetx", frm, to, value, fee=fee, estimate_fee_blocks=estimate_fee_blocks,
                               change_addr=change_addr)

    def preparesignedmultirecipienttx(self, privkey: PrivateKeySignAllType, frm: str, outs: List[TxOut],
                                      change_addr: str = None, fee: int = None, estimate_fee_blocks: int = 6) -> Tx:
        return self._run_async("preparesignedmultirecipienttx", privkey, frm, outs,
                               change_addr=change_addr, fee=fee, estimate_fee_blocks=estimate_fee_blocks)

    def preparesignedtx(self, privkey: PrivateKeySignAllType, frm: str, to: str, value: int, change_addr: str = None,
                        fee: int = None, estimate_fee_blocks: int = 6) -> Tx:
        return self._run_async("preparesignedtx", privkey, frm, to, value,
                               change_addr=change_addr, fee=fee, estimate_fee_blocks=estimate_fee_blocks)

    def send_to_multiple_receivers_tx(self, privkey: PrivateKeySignAllType, addr: str, outs: List[TxOut],
                                      change_addr: str = None, fee: int = None, estimate_fee_blocks: int = 6):
        return self._run_async("send_to_multiple_receivers_tx", privkey, addr, outs,
                               change_addr=change_addr, fee=fee, estimate_fee_blocks=estimate_fee_blocks)

    def send(self, privkey: PrivateKeySignAllType, frm: str, to: str, value: int, change_addr: str = None,
             fee: int = None, estimate_fee_blocks: int = 6):
        return self._run_async("send", privkey, frm, to, value,
                               change_addr=change_addr, fee=fee, estimate_fee_blocks=estimate_fee_blocks)

    def inspect(self, tx: Union[str, Tx]) -> TXInspectType:
        return self._run_async("inspect", tx)

    def wallet(self, seed: str, passphrase: str = None, **kwargs) -> HDWallet:
        return self._async_coin.wallet(seed, passphrase=passphrase, **kwargs)

    def watch_wallet(self, xpub: str, **kwargs) -> HDWallet:
        return self._async_coin.watch_wallet(xpub, **kwargs)

    def p2wpkh_p2sh_wallet(self, seed: str, passphrase: str = None, **kwargs) -> HDWallet:
        return self._async_coin.p2wpkh_p2sh_wallet(seed, passphrase=passphrase, **kwargs)

    def watch_p2wpkh_p2sh_wallet(self, xpub: str, **kwargs) -> HDWallet:
        return self._async_coin.watch_p2wpkh_p2sh_wallet(xpub, **kwargs)

    def p2wpkh_wallet(self, seed: str, passphrase: str = None, **kwargs) -> HDWallet:
        return self._async_coin.p2wpkh_wallet(seed, passphrase=passphrase, **kwargs)

    def watch_p2wpkh_wallet(self, xpub: str, **kwargs) -> HDWallet:
        return self._async_coin.watch_p2wpkh_wallet(xpub, **kwargs)

    def electrum_wallet(self, seed: str, passphrase: str = None, **kwargs) -> HDWallet:
        return self._async_coin.electrum_wallet(seed, passphrase=passphrase, **kwargs)

    def watch_electrum_wallet(self, xpub: str, **kwargs) -> HDWallet:
        return self._async_coin.watch_electrum_wallet(xpub, **kwargs)

    def watch_electrum_p2wpkh_wallet(self, xpub: str, **kwargs) -> HDWallet:
        return self._async_coin.watch_electrum_p2wpkh_wallet(xpub, **kwargs)

    def is_cash_address(self, addr: str) -> bool:
        return self._async_coin.is_cash_address(addr)

    def scripthash_to_cash_addr(self, scripthash: bytes) -> str:
        return self._async_coin.scripthash_to_cash_addr(scripthash)

    def hash_to_cash_addr(self, pub_hash: AnyStr) -> str:
        return self._async_coin.hash_to_cash_addr(pub_hash)

    def pub_to_cash_address(self, pubkey: str) -> str:
        return self._async_coin.pub_to_cash_address(pubkey)

    def privtocashaddress(self, privkey: PrivkeyType) -> str:
        return self._async_coin.privtocashaddress(privkey)

    def legacy_addr_to_cash_address(self, addr: str) -> str:
        return self._async_coin.legacy_addr_to_cash_address(addr)

    def cash_address_to_legacy_addr(self, addr: str) -> str:
        return self._async_coin.cash_address_to_legacy_addr(addr)

    def mk_multsig_cash_address(self, *args: str, num_required: int = None) -> Tuple[str, str]:
        return self._async_coin.mk_multsig_cash_address(*args, num_required=num_required)

    def calculate_fee(self, tx: Tx) -> int:
        return self._run_async("calculate_fee", tx)



import unittest

from cryptos import coins_async
from cryptos.main import script_to_scripthash
from cryptos.types import TxInput, Tx
from cryptos.testing.testcases_async import BaseAsyncCoinTestCase
from cryptos.electrumx_client.types import ElectrumXTx, ElectrumXMultiBalanceResponse
from typing import List, Type
from unittest import mock


class TestBitcoin(BaseAsyncCoinTestCase):
    name: str = "Bitcoin"
    coin: Type[coins_async.BaseCoin] = coins_async.Bitcoin
    addresses: List[str] = ["1MhTCMUjM1TEQ7RSwoCMVZy7ARa9aAP82Z",
                            "18DEbpqpdmfNaosxwQhCNHDAChZYCNG836",
                            "175MvCJkNZT3zSdCntXj9vK7L6XKDWjLnD"]
    segwit_addresses: List[str] = ["3FWfXAgccKXVSVcmEF3RRf2KacBSQ6rcix",
                                   "3NFyu5P3C9PvWUB5ekvmRgYgniNEE1n5za",
                                   "32UT11rEHGqDnMZxM1VGfz4dLMxNrJWnu3"]
    native_segwit_addresses: List[str] = ["bc1q95cgql39zvtc57g4vn8ytzmlvtt43sknztmu82",
                                          "bc1qfuvnn87p787z7nqv9seu4e8fqel83yaczcl63s",
                                          "bc1qst3pkm860tjt9y70ugnaluqyqnfa7h54nslppf"]
    native_segwit_multisig_addresses: List[str] = ["bc1q7e42a8gmgp5d7kw8myrruj5qnxp8edp7v5y0y95hrjz6t530uehqp6ey99",
                                                   "bc1qu7fz4uku8kh6tg7qghj7rnf88g858lal258gzfu85kx7vc5h0qpsnw4xfw"]
    multisig_addresses: List[str] = ["35D72hVBzYXqNkyN63z28FHmSyPKuJh9Q2", "32tuh24PcKWQWfWitfp9NVhRuYjDKG7vCH"]
    privkeys: List[str] = ["098ddf01ebb71ead01fc52cb4ad1f5cafffb5f2d052dd233b3cad18e255e1db1",
                           "0861e1bb62504f5e9f03b59308005a6f2c12c34df108c6f7c52e5e712a08e91401",
                           "c396c62dfdc529645b822dc4eaa7b9ddc97dd8424de09ca19decce61e6732f71"]
    privkey_standard_wifs: List[str] = ['5HtVdWeLBUpSyqqSnnJoQnCiuc33Kg86i8V1gc1tKK13tw1Cqrg',
                                        "KwW1FKxkfefDyVStxvKH9qCCb9qaiFXBFZUy2mPLvTMap2f5YaXR",
                                        "5KJRe1vYXbE4hhvNjJjPX6iS1tqpksNKHChrQjzyYVDgh9Z8H5o"]
    privkey_segwit_wifs: List[str] = ["LEhYWUZa1ZchcvaMwDmMSJhiU43JvGDnXRGu8iF6UmxeLZ9MvbRp",
                                      "LEfGPegZ1xVVvXUS5574kBPB62EtP64daSjE9g5XuQs1Nh2hh3u3",
                                      "LLwAtho8WKtwHhe81t3szCQqupDoTTrY1PuaChV5Yw4QkRnUhx4Y"]
    privkey_native_segwit_wifs: List[str] = ["L67uwKCfqQhZeQZbP9NTe8cEDcqf5qx4MyemaFtzzJCw4DvrFkVK",
                                             "L65dpVKeqoaMx1TfWziAx1Hgqb3EYfnuR176bDjSQw7J6MrvP2GZ",
                                             "LCMYKYSELAyoKBdMToezC2KMfP29d3aoqxHSeF8z4TJhU6gYDEQV"]
    fee: int = 54400
    max_fee: int = fee
    testnet: bool = False
    balance: ElectrumXMultiBalanceResponse = {'confirmed': 16341002035, 'unconfirmed': 0}
    balances: List[ElectrumXMultiBalanceResponse] = [
        {'confirmed': 16341002035, 'unconfirmed': 0, 'address': '12gK1NsNhzrRxs2kGKSjXhA1bhd8vyyWMR'},
        {'confirmed': 8000100547, 'unconfirmed': 0, 'address': '1A7hMTCfHbQJ1RAtBAVNcUtVsh8i8yFdmT'}]
    history: List[ElectrumXTx] = [{'tx_hash': 'b489a0e8a99daad4d1a85992d9e373a87463a95109a5c56f4e4827f4e5a1af34', 'height': 114743},
                {'tx_hash': 'f5e0c14b7d1f95d245d990ac6bb9ccf28d7f80f721f8133cd6ed34f9c8d13f0f', 'height': 116768},
               {'height': 547011,
                'tx_hash': 'e66ebcd04e86196b59e8ff54c071fb82d055c40cbd7314309088e6c2b5658a0a'},
               {'height': 621585,
                'tx_hash': '655620ced5f7f6ff0edcb930ab787f1e61a0872ce8d318a94ff884a9c7e81808'},
               {'height': 651450,
                'tx_hash': '04eac4e98bc74b344d85bd1f008d227d8a7715224f5d1af0929810c08fd7fed2'}]
    histories: List[ElectrumXTx] = [{'address': '12gK1NsNhzrRxs2kGKSjXhA1bhd8vyyWMR',
      'height': 651450,
      'tx_hash': '04eac4e98bc74b344d85bd1f008d227d8a7715224f5d1af0929810c08fd7fed2'},
     {'address': '12gK1NsNhzrRxs2kGKSjXhA1bhd8vyyWMR',
      'height': 621585,
      'tx_hash': '655620ced5f7f6ff0edcb930ab787f1e61a0872ce8d318a94ff884a9c7e81808'},
     {'address': '1A7hMTCfHbQJ1RAtBAVNcUtVsh8i8yFdmT',
      'height': 248365,
      'tx_hash': 'a146923df9579f7c7b9a8f5ddf27e230e8d838117379bdf6b57113ce31bf52e0'},
     {'address': '12gK1NsNhzrRxs2kGKSjXhA1bhd8vyyWMR',
      'height': 114743,
      'tx_hash': 'b489a0e8a99daad4d1a85992d9e373a87463a95109a5c56f4e4827f4e5a1af34'},
     {'address': '1A7hMTCfHbQJ1RAtBAVNcUtVsh8i8yFdmT',
      'height': 659070,
      'tx_hash': 'cc0b5794055a0f84682c29cd6fee42d595cf7624045cf2920768694f36ca8dca'},
     {'address': '12gK1NsNhzrRxs2kGKSjXhA1bhd8vyyWMR',
      'height': 547011,
      'tx_hash': 'e66ebcd04e86196b59e8ff54c071fb82d055c40cbd7314309088e6c2b5658a0a'},
     {'address': '12gK1NsNhzrRxs2kGKSjXhA1bhd8vyyWMR',
      'height': 116768,
      'tx_hash': 'f5e0c14b7d1f95d245d990ac6bb9ccf28d7f80f721f8133cd6ed34f9c8d13f0f'},
     {'address': '1A7hMTCfHbQJ1RAtBAVNcUtVsh8i8yFdmT',
      'height': 187296,
      'tx_hash': 'fd232fe21b6ad7f096f3012e935467a7f2177258cdcd07c748502a5b1f31ccd5'}]
    unspent_addresses: List[str] = ["12gK1NsNhzrRxs2kGKSjXhA1bhd8vyyWMR", "1A7hMTCfHbQJ1RAtBAVNcUtVsh8i8yFdmT"]
    unspent: List[ElectrumXTx] = [
        {'tx_hash': 'b489a0e8a99daad4d1a85992d9e373a87463a95109a5c56f4e4827f4e5a1af34', 'tx_pos': 1, 'height': 114743,
         'value': 5000000, 'address': '12gK1NsNhzrRxs2kGKSjXhA1bhd8vyyWMR'},
        {'tx_hash': 'f5e0c14b7d1f95d245d990ac6bb9ccf28d7f80f721f8133cd6ed34f9c8d13f0f', 'tx_pos': 1, 'height': 116768,
         'value': 16336000000, 'address': '12gK1NsNhzrRxs2kGKSjXhA1bhd8vyyWMR'},
        {'tx_hash': 'e66ebcd04e86196b59e8ff54c071fb82d055c40cbd7314309088e6c2b5658a0a', 'tx_pos': 1974, 'height': 547011,
         'value': 888, 'address': '12gK1NsNhzrRxs2kGKSjXhA1bhd8vyyWMR'},
        {'tx_hash': '655620ced5f7f6ff0edcb930ab787f1e61a0872ce8d318a94ff884a9c7e81808', 'tx_pos': 1, 'height': 621585,
         'value': 600, 'address': '12gK1NsNhzrRxs2kGKSjXhA1bhd8vyyWMR'},
        {'tx_hash': '04eac4e98bc74b344d85bd1f008d227d8a7715224f5d1af0929810c08fd7fed2', 'tx_pos': 531, 'height': 651450,
         'value': 547, 'address': '12gK1NsNhzrRxs2kGKSjXhA1bhd8vyyWMR'}
    ]
    unspents: List[ElectrumXTx] = unspent + [
        {'height': 659070, 'tx_hash': 'cc0b5794055a0f84682c29cd6fee42d595cf7624045cf2920768694f36ca8dca', 'tx_pos': 40,
         'value': 547, 'address': '1A7hMTCfHbQJ1RAtBAVNcUtVsh8i8yFdmT'},
        {'tx_hash': 'fd232fe21b6ad7f096f3012e935467a7f2177258cdcd07c748502a5b1f31ccd5', 'tx_pos': 0, 'height': 187296,
         'value': 8000000000, 'address': '1A7hMTCfHbQJ1RAtBAVNcUtVsh8i8yFdmT'},
        {'tx_hash': 'a146923df9579f7c7b9a8f5ddf27e230e8d838117379bdf6b57113ce31bf52e0', 'tx_pos': 41, 'height': 248365,
         'value': 100000, 'address': '1A7hMTCfHbQJ1RAtBAVNcUtVsh8i8yFdmT'}]

    min_latest_height: int = 503351
    txid: str = "fd3c66b9c981a3ccc40ae0f631f45286e7b31cf6d9afa1acaf8be1261f133690"
    txheight: int = 509045
    block_hash: str = "000000000000000000103a149fa9a449e5b4840fb18d22d5458eb650e1098ea9"
    txinputs: List[TxInput] = [{"output": "7a905da948f1e174c43c6f41b0a0ee338119191de7b92bd1ca3c79f899e5d583:1", 'value': 1000000},
                               {"output": "da1ad82b777c51105d3a24cef253e0301dd08153115013a49e0edf69fd7cdadf:1", 'value': 100000}]
    tx: Tx = {'ins': [{'tx_hash': '7a905da948f1e174c43c6f41b0a0ee338119191de7b92bd1ca3c79f899e5d583', 'tx_pos': 1,
                       'script': '483045022076bf3b0edd6c9cdd35fb30d77d780f1d752e959242b2bbd58123617b8db350a6022100a602b91002b9c6c078a7513f72e1d7ccbfa3aa6f1261706b3110db00b1205ae4014104fafb576fcaf43a773ee1e34c5a76ab1f4fe1a7dc23256dd7a4525092537fc11686227d495dff710a291e7e9a6bf474a968158c56882b153e4b2e17bc584ec3cc',
                       'sequence': 4294967295},
                   {'tx_hash': 'da1ad82b777c51105d3a24cef253e0301dd08153115013a49e0edf69fd7cdadf', 'tx_pos': 1,
                    'script': '493046022100aecef1b98cf1cead7daadfb538c4808e71c9ef0c1ecec04af64fb1fdcffa7afb022100ec1070f8dea90f9ef6d86ebf251a63a01eae48ff840e0aacce899775b2dd16c6014104d2eeecdff2d0fd3d19f07928689f2aed33f1298f7493f2ca77b3607b545a8b2a91af48c27bc949da72f6ef38412c95bdcf6618486207bb92cd9aa75cae2c116d',
                    'sequence': 4294967295}],
              'outs': [{'value': 100000, 'script': '76a91462e907b15cbf27d5425399ebf6f0fb50ebb88f1888ac'}], 'version': 1,
              'locktime': 0}

    def test_standard_wif_ok(self):
        self.assertStandardWifOK()

    def test_p2wpkh_p2sh_wif_ok(self):
        self.assertP2WPKH_P2SH_WifOK()

    def test_p2wpkh_wif_ok(self):
        self.assertP2WPKH_WIFOK()

    def test_p2pw_p2sh_address_ok(self):
        pubs = ["02ba64c5ace172273914d2964e1c97bf46bfadc3efa122e19f06bfd04a9dd290dd",
                "02e1dff2d67d6f35d8c8e1983cad52649aa1355793202d948e5015da49ac848481",
                "03bd517db94bce99d6f61f4439ad3c8671a5dc6504615aa101142be2d30a41c01e"]
        expected_witness_script = "522102ba64c5ace172273914d2964e1c97bf46bfadc3efa122e19f06bfd04a9dd290dd2102e1dff2d67d6f35d8c8e1983cad52649aa1355793202d948e5015da49ac8484812103bd517db94bce99d6f61f4439ad3c8671a5dc6504615aa101142be2d30a41c01e53ae"
        expected_address = "bc1qvyh0kyxxtv02z4uwej509df6mvgspvfe0u4ekkwkxxpt2rz6t8dqfxljp2"
        num_required = 2
        script, address = self._coin.mk_multsig_segwit_address(*pubs, num_required=num_required)
        self.assertEqual(script, expected_witness_script)
        self.assertEqual(address, expected_address)
        expected_script_for_scripthash = "0020612efb10c65b1ea1578ecca8f2b53adb1100b1397f2b9b59d63182b50c5a59da"
        expected_scripthash = "fc1546b5baf99c4eb4d04c66475a5560130165ceb0ac24a516065b535fd3dcbb"
        script = self._coin.addrtoscript(address)
        scripthash = script_to_scripthash(script)
        self.assertEqual(expected_script_for_scripthash, script)
        self.assertEqual(expected_scripthash, scripthash)

    async def test_balance(self):
        await self.assertBalanceOK()

    async def test_balances(self):
        await self.assertBalancesOK()

    async def test_unspent(self):
        await self.assertUnspentOK()

    async def test_unspents(self):
        await self.assertUnspentsOK()

    async def test_merkle_proof(self):
        await self.assertMerkleProofOK()

    async def test_history(self):
        await self.assertHistoryOK()

    async def test_histories(self):
        await self.assertHistoriesOK()

    async def test_block_header(self):
        await self.assertBlockHeaderOK()

    async def test_block_headers(self):
        await self.assertBlockHeadersOK()

    @unittest.skip("Intermittent failures")
    async def test_gettx(self):
        await self.assertGetTXOK()

    async def test_balance_merkle_proven(self):
        await self.assertBalanceMerkleProvenOK()

    async def test_balances_merkle_proven(self):
        await self.assertBalancesMerkleProvenOK()

    async def test_getverbosetx(self):
        await self.assertGetVerboseTXOK()

    async def test_gettxs(self):
        await self.assertTxsOK()

    async def test_transaction(self):
        with mock.patch('cryptos.electrumx_client.client.NotificationSession.send_request',
                        side_effect=self.mock_electrumx_send_request):
            await self.assertTransactionOK("b69882d143ef353878f0fe6b0579ffddf01309fd7a9ccdf4247fdd1e306c55f4")

    async def test_transaction_p2pk(self):
        with mock.patch('cryptos.electrumx_client.client.NotificationSession.send_request',
                        side_effect=self.mock_electrumx_send_request):
            await self.assertTransactionToPKOK(
                "e84595d05a8e6b1227b8591494d220f943a5ae6a85f33da6d444a86559cac4f1")

    async def test_transaction_segwit(self):
        with mock.patch('cryptos.electrumx_client.client.NotificationSession.send_request',
                        side_effect=self.mock_electrumx_send_request):
            await self.assertSegwitTransactionOK(
                "efe90a8e6b7154cbac3db9efb5ee92edca8107f3c7f7308eb79c9a81771f2a3a")

    async def test_transaction_native_segwit(self):
        with mock.patch('cryptos.electrumx_client.client.NotificationSession.send_request',
                        side_effect=self.mock_electrumx_send_request):
            await self.assertNativeSegwitTransactionOK(
                "ca7212dd4f46a624b681a775710dbff758fd0e31eb8c73ae0aafec2e34c5b405")

    async def test_transaction_mixed_segwit(self):
        with mock.patch('cryptos.electrumx_client.client.NotificationSession.send_request',
                        side_effect=self.mock_electrumx_send_request):
            await self.assertMixedSegwitTransactionOK("578d8ea64cf1e1608c7064ab913ca911e62e16d341d5ca7fbabd2f92951e889e")

    async def test_transaction_multisig(self):
        with mock.patch('cryptos.electrumx_client.client.NotificationSession.send_request',
                        side_effect=self.mock_electrumx_send_request):
            await self.assertMultiSigTransactionOK("8a112bb83ac5a8a12ea6183acb080785921ab8cd2689163527300dba173958e7")

    async def test_transaction_native_segwit_multisig(self):
        with mock.patch('cryptos.electrumx_client.client.NotificationSession.send_request',
                        side_effect=self.mock_electrumx_send_request):
            await self.assertNativeSegwitMultiSigTransactionOK("b2dfcf79c5308f179e33136e5d6faa6adcab2d63d33e4a4172f91e616241a336")

    async def test_sendmulti_recipient_tx(self):
        with mock.patch('cryptos.electrumx_client.client.NotificationSession.send_request',
                        side_effect=self.mock_electrumx_send_request):
            await self.assertSendMultiRecipientsTXOK("29ce6ce80bf5f381eae1b27049e6b15d2a6316664b1f8ffc3404293b2a4b56e2")

    async def test_send(self):
        with mock.patch('cryptos.electrumx_client.client.NotificationSession.send_request',
                        side_effect=self.mock_electrumx_send_request):
            await self.assertSendOK("f02bbffef56329eae1d48dd2471563e82ce10ed644db34a3b2bf19a28ecd80ed")

    async def test_subscribe_block_headers(self):
        await self.assertSubscribeBlockHeadersOK()

    async def test_subscribe_block_headers_sync(self):
        await self.assertSubscribeBlockHeadersSyncCallbackOK()

    async def test_latest_block(self):
        await self.assertLatestBlockOK()

    async def test_confirmations(self):
        await self.assertConfirmationsOK()



import asyncio
import json
import itertools
import threading
from concurrent.futures import Future
from ssl import create_default_context, Purpose
from typing import Dict, Union


scripthash_status1 = "e1969d52d5c94cdc9f3839ef720eec70282ce4c76d3634d2bdf138e24b223dc8"
scripthash_status2 = "e1969d52d5c94cdc9f3839ef720eec70282ce4c76d3634d2bdf138e24b223d44"
scripthash_status3 = "e1969d52d5c94cdc9f3839ef720eec70282ce4c76d3634d2bdf138e24b223dxy"

first_conn_scripthash = [scripthash_status1, scripthash_status2]
second_conn_scripthash = [scripthash_status2, scripthash_status3]


block1 = {'height': 2350325, 'hex': "0000ff3f7586812b8a8677342ceef85916c2667b63468a8d19d0604c2e000000000000005292d8eba79db851be100996f48147df69386b43bf7fcb5e3361cf46f9ea8ed8a3214463ffff001d51c56337"}
block2 = {'height': 2350326, 'hex': "00004a2920c2d8311e12d3e35b8da48ad29b1254e0a0d2be1623d717f69000000000000001aaf14e207eea7f36cdc1cf92a8d43a5db2ac1ce22925e104ccedca3b5d2d26892544630194331933c10227"}
block3 = {'height': 2350327, 'hex': "0000552920c2d8311e12d3e35b8da48ad29b1254e0a0d2be1623d717f69000000000000001aaf14e207eea7f36cdc1cf92a8d43a5db2ac1ce22925e104ccedca3b5d2d26892544630194331933c10227"}

first_conn_blocks = [block1, block2]
second_conn_blocks = [block2, block3]


cycles = {}


def reset_cycles() -> None:
    cycles['scripthash'] = itertools.cycle([first_conn_scripthash, second_conn_scripthash])
    cycles['block'] = itertools.cycle([first_conn_blocks, second_conn_blocks])


reset_cycles()


async def send_message(writer, data: Dict[str, Union[int, str, dict]]) -> None:
    data['jsonrpc'] = "2.0"

    out_message = (json.dumps(data) + "\n").encode()

    print('sending', out_message)

    writer.write(out_message)
    await writer.drain()


async def handle_rpc(reader: asyncio.StreamReader, writer: asyncio.StreamWriter):
    while not writer.transport.is_closing():
        print('Server Reading data')
        msg = await reader.readline()
        print('Server received data', msg)
        if msg:
            data = msg.decode()

            data = json.loads(data)
            print("Server json data", data)
            msg_id = data['id']
            method = data['method']
            params = data.get('params', [])
            print("Server received method", method)
            if method == "server.version":
                response = {"result": ["ElectrumXMock 1.16.0", "1.4"], "id": msg_id}
                await send_message(writer, response)
            elif method == "server.ping":
                response = {"result": None, "id": msg_id}
                await send_message(writer, response)
            elif method == "blockchain.scripthash.subscribe":
                scripthash_status = next(cycles['scripthash'])
                response = {
                    "result": scripthash_status[0],
                    "id": msg_id
                }
                await send_message(writer, response)
                await asyncio.sleep(2)
                response = {
                    "method": method,
                    "params": params + [scripthash_status[1]]
                }
                await send_message(writer, response)
            elif method == "blockchain.headers.subscribe":
                blocks = next(cycles['block'])
                response = {
                    "result": blocks[0],
                    "id": msg_id}
                await send_message(writer, response)
                await asyncio.sleep(2)
                response = {
                    "method": method,
                    "params": [blocks[1]]
                }
                await send_message(writer, response)
            else:
                response = {"error": {"code": -32601, "message": f"Method {method} is not supported by this mock server", "id": msg_id}}
                await send_message(writer, response)
        else:
            print('Closing writer')
            writer.close()
    print('Server stopped while loop')


async def get_server(fut: Union[asyncio.Future, Future], host: str = "127.0.0.1", ssl: bool = False):
    port = 44445 if ssl else 44444
    context = create_default_context(purpose=Purpose.SERVER_AUTH) if ssl else None
    server = await asyncio.start_server(
        handle_rpc, host, port, ssl=context)

    sock = server.sockets[0]
    listening_on = sock.getsockname()

    asyncio.get_running_loop().call_soon(fut.set_result, listening_on)

    return server


async def run_server(fut: asyncio.Future, host: str = "127.0.0.1", ssl: bool = False):
    server = await get_server(fut, host=host, ssl=ssl)
    async with server:
        await server.serve_forever()


async def run_server_until_cancelled(fut: Future, stop_fut: Future, host: str = "127.0.0.1", ssl: bool = False):
    stop_fut_async = asyncio.wrap_future(stop_fut)
    server = await get_server(fut, host=host, ssl=ssl)
    async with server:
        await asyncio.wait([asyncio.create_task(server.serve_forever()), stop_fut_async], return_when=asyncio.FIRST_COMPLETED)


def run_in_loop(fut: Future, stop_fut: Future, host: str = '127.0.0.1', ssl: bool = False):
    asyncio.run(run_server_until_cancelled(fut, stop_fut, host=host, ssl=ssl))


def run_server_in_thread(stop_fut: Future, host: str = '127.0.0.1', ssl: bool = False):
    fut = Future()
    thread = threading.Thread(target=run_in_loop, args=(fut, stop_fut), kwargs={'host': host, 'ssl': ssl}, daemon=True)
    thread.start()
    return fut.result(10)



import unittest

from cryptos import coins_async
from cryptos.testing.testcases_async import BaseAsyncCoinTestCase
from cryptos.electrumx_client.types import ElectrumXTx, ElectrumXMultiBalanceResponse
from cryptos.types import TxInput
from typing import List, Type
from unittest import mock


class TestLitecoin(BaseAsyncCoinTestCase):
    name: str = "Litecoin"
    coin: Type[coins_async.BaseCoin] = coins_async.Litecoin
    addresses: List[str] = ["LfvQTZnZRfhHev7c7wBemb2sNdwRf4eeEC",
                            "LSSBs39eiRuRqca87YgVeJGvQuvpNwaVkr",
                            "LRJKBQcaTDh7FFKMy2X2RwNsYJtbJBvfvc"]
    segwit_addresses: List[str] = ["MMioq46aZSNvEztfL82mFJGiuJmtRVTFAp",
                                   "MUU8Cxo19GFMJySykdv7FKo67QxgEHzvPG",
                                   "M8gbJuGCEPgearqrStUcVdK2f4YpqCmp9d"]
    native_segwit_addresses: List[str] = ["ltc1q95cgql39zvtc57g4vn8ytzmlvtt43sknxhpcl6",
                                          "ltc1qfuvnn87p787z7nqv9seu4e8fqel83yacxy97fq",
                                          "ltc1qst3pkm860tjt9y70ugnaluqyqnfa7h54hv99ee"]
    multisig_addresses: List[str] = ["MBRFLau9wfPGBGFGBvyMwtYAmfymspELC4", "M973zuUMZSMqKAnczYoVC8wqEFKfHMjANu"]
    native_segwit_multisig_addresses: List[str] = [
        "ltc1q7e42a8gmgp5d7kw8myrruj5qnxp8edp7v5y0y95hrjz6t530uehqz7h5lq",
        "ltc1qu7fz4uku8kh6tg7qghj7rnf88g858lal258gzfu85kx7vc5h0qpss2mknt"
    ]
    privkeys: List[str] = ["098ddf01ebb71ead01fc52cb4ad1f5cafffb5f2d052dd233b3cad18e255e1db1",
                           "0861e1bb62504f5e9f03b59308005a6f2c12c34df108c6f7c52e5e712a08e91401",
                           "c396c62dfdc529645b822dc4eaa7b9ddc97dd8424de09ca19decce61e6732f71"]
    privkey_standard_wifs: List[str] = ['6uCE6eBs5uHKTDjJJc6mCAyts5bWXUa8UotBPo2v2mKfakiBJnv',
                                        "T3LGh5Fw52dpkL5mWZG9NBjaY1UtnLY54mPDta1tVRXkKvEGabgc",
                                        "6vcA79U5S1gwB5pEF8XMJVVbyNQHxfpM3t727w21FwYJNzcDXkb"]
    privkey_segwit_wifs: List[str] = ["TLXoxDrkQwbJPmDEUriDefF6QugczMEgLdB9zWse3k8orSi53T7e",   # Not supported in eleectrum
                                      "TLVXqPyjRLU6hN7Jci3vxXvZ2stCTB5XPedV1Ui5UP3AtaYfaxw2",
                                      "TSmSLT6JuhsY4YGzZWzkCYxDrfs7XYsRpboq4W7d7uEaGKLwHQfc"]
    privkey_native_segwit_wifs: List[str] = ["TBxBP4VrEngARFCTvnKKrV9cAUUy9vxxBBZ2S4XYZGP6a7WtRR8a",    # Not recognised in electrum
                                             "TBuuGEcqFBYxir6Y4df3AMq4nSgYckooED1MT2MyyuHTcFPhEP7W",
                                             "TJBomHjQjYxQ62GE1SbrQNrjcEfTh8bhfABhW3mXdRUryzB2UR2g"]
    fee: int = 54400
    max_fee: int = fee
    testnet: bool = False

    unspent_addresses: List[str] = ["LSdTvMHRm8sScqwCi6x9wzYQae8JeZhx6y"]
    balance: ElectrumXMultiBalanceResponse = {'confirmed': 83515295, 'unconfirmed': 0}
    balances: List[ElectrumXMultiBalanceResponse] = [{'address': unspent_addresses[0]} | dict(balance)]
    unspent: List[ElectrumXTx] = [{'address': 'LSdTvMHRm8sScqwCi6x9wzYQae8JeZhx6y',
                                  'height': 1495972,
                                  'tx_hash': '6cf532663cd14013aa6ccb394f86d64aa48fce4d6aa8a175f9b75ea486465ca9',
                                  'tx_pos': 0,
                                  'value': 1984},
                                 {'address': 'LSdTvMHRm8sScqwCi6x9wzYQae8JeZhx6y',
                                  'height': 1509957,
                                  'tx_hash': '4987fcebf1643d6a254e6352b72867de1c64a0f5b9e6e1e83c103be419ed9914',
                                  'tx_pos': 1,
                                  'value': 10000917},
                                 {'address': 'LSdTvMHRm8sScqwCi6x9wzYQae8JeZhx6y',
                                  'height': 1514616,
                                  'tx_hash': '107cc3d3bf8f1977f9c60b80d02c6e44b1e272a8ebe0d43f712762113bb47c43',
                                  'tx_pos': 0,
                                  'value': 262765},
                                 {'address': 'LSdTvMHRm8sScqwCi6x9wzYQae8JeZhx6y',
                                  'height': 1710946,
                                  'tx_hash': '5acea20cf6d767ad4e4c32b4c92354a3545ef147de75494a00b74ad0998b2fa7',
                                  'tx_pos': 0,
                                  'value': 9977400},
                                 {'address': 'LSdTvMHRm8sScqwCi6x9wzYQae8JeZhx6y',
                                  'height': 1762963,
                                  'tx_hash': '7f32b287c6264965f8cac174d0b106163572516cf573907a54ad15ce2a47dfa0',
                                  'tx_pos': 0,
                                  'value': 7507184},
                                 {'address': 'LSdTvMHRm8sScqwCi6x9wzYQae8JeZhx6y',
                                  'height': 1770983,
                                  'tx_hash': 'cdebcffb1622fe8ea511f2f6ed9059b6b3eca1dba547d6386d21564a15571b13',
                                  'tx_pos': 863,
                                  'value': 23045},
                                 {'address': 'LSdTvMHRm8sScqwCi6x9wzYQae8JeZhx6y',
                                  'height': 1800218,
                                  'tx_hash': '68f98cd415c0586c917e4ba6c78747e7281adca09837048ff8fa211cab9326cd',
                                  'tx_pos': 1,
                                  'value': 54059898},
                                 {'address': 'LSdTvMHRm8sScqwCi6x9wzYQae8JeZhx6y',
                                  'height': 2192455,
                                  'tx_hash': '8391ba465fe818e7041a2316e508731b6e20bf9488c0b7579bfd7de46c6e9ab3',
                                  'tx_pos': 0,
                                  'value': 360933},
                                 {'address': 'LSdTvMHRm8sScqwCi6x9wzYQae8JeZhx6y',
                                  'height': 2262951,
                                  'tx_hash': '43eb70acc4c8c6bdbfabdf76cdbf0a6a969b82051514d105f8c00dff23e715c6',
                                  'tx_pos': 0,
                                  'value': 200000},
                                 {'address': 'LSdTvMHRm8sScqwCi6x9wzYQae8JeZhx6y',
                                  'height': 2352771,
                                  'tx_hash': '7158c6f05ef2b06a40cd00b64647da6735d2eb3908f958fcbc5d0584f34be460',
                                  'tx_pos': 0,
                                  'value': 1121169}]
    history: List[ElectrumXTx] = [{'height': 1495972,
                                   'tx_hash': '6cf532663cd14013aa6ccb394f86d64aa48fce4d6aa8a175f9b75ea486465ca9'},
                                  {'height': 1509957,
                                   'tx_hash': '4987fcebf1643d6a254e6352b72867de1c64a0f5b9e6e1e83c103be419ed9914'},
                                  {'height': 1514616,
                                   'tx_hash': '107cc3d3bf8f1977f9c60b80d02c6e44b1e272a8ebe0d43f712762113bb47c43'},
                                  {'height': 1710946,
                                   'tx_hash': '5acea20cf6d767ad4e4c32b4c92354a3545ef147de75494a00b74ad0998b2fa7'},
                                  {'height': 1762963,
                                   'tx_hash': '7f32b287c6264965f8cac174d0b106163572516cf573907a54ad15ce2a47dfa0'},
                                  {'height': 1770983,
                                   'tx_hash': 'cdebcffb1622fe8ea511f2f6ed9059b6b3eca1dba547d6386d21564a15571b13'},
                                  {'height': 1800218,
                                   'tx_hash': '68f98cd415c0586c917e4ba6c78747e7281adca09837048ff8fa211cab9326cd'},
                                  {'height': 2192455,
                                   'tx_hash': '8391ba465fe818e7041a2316e508731b6e20bf9488c0b7579bfd7de46c6e9ab3'},
                                  {'height': 2262951,
                                   'tx_hash': '43eb70acc4c8c6bdbfabdf76cdbf0a6a969b82051514d105f8c00dff23e715c6'},
                                  {'height': 2352771,
                                   'tx_hash': '7158c6f05ef2b06a40cd00b64647da6735d2eb3908f958fcbc5d0584f34be460'}]
    histories: List[ElectrumXTx] = [dict(h) | {'address': "LSdTvMHRm8sScqwCi6x9wzYQae8JeZhx6y"} for h in history]
    unspents: List[ElectrumXTx] = unspent
    min_latest_height: int = 2360220
    txheight: int = 509045
    block_hash: str = "3bfd5d7f95a28e1af34cc50a36aee92c4aa4c317d4d0f7795034f7a8fc74b6c9"
    txid: str = "0c2d49e00dd1372a7219fbc4378611b39f54790bbd597b4c29517f0d93c9faa2"
    txinputs: List[TxInput] = [{'output': 'b3105972beef05e88cf112fd9718d32c270773462d62e4659dc9b4a2baafc038:0', 'value': 1157763509}]
    raw_tx: str = "020000000138c0afbaa2b4c99d65e4622d467307272cd31897fd12f18ce805efbe725910b3000000006b483045022100b9dfcf3d1cc384797db57f14adf17716c6dc8d47f7733a3ebceb233ed9d18cc202205f44701f5f371404026a165ffe2dbd405853ca838c28710aa64bb3038102f5de012102158315db717fbda2d709f41a913fdfc384f269884dc0915bd0a347bc3cf3750bffffffff02ffe30942000000001976a914bb2c3d60bfaf9d3b87fc9f225e8a925200d4832388ac968cf002000000001976a91444baf967037464e3c3f5bfe7ae2c98385aca39c188ac00000000"

    def test_standard_wif_ok(self):
        self.assertStandardWifOK()

    def test_p2wpkh_p2sh_wif_ok(self):
        self.assertP2WPKH_P2SH_WifOK()

    def test_p2wpkh_wif_ok(self):
        self.assertP2WPKH_WIFOK()

    async def test_balance(self):
        await self.assertBalanceOK()

    async def test_balances(self):
        await self.assertBalancesOK()

    async def test_unspent(self):
        await self.assertUnspentOK()

    async def test_unspents(self):
        await self.assertUnspentsOK()

    async def test_merkle_proof(self):
        await self.assertMerkleProofOK()

    async def test_history(self):
        await self.assertHistoryOK()

    async def test_histories(self):
        await self.assertHistoriesOK()

    async def test_block_header(self):
        await self.assertBlockHeaderOK()

    async def test_block_headers(self):
        await self.assertBlockHeadersOK()

    async def test_gettx(self):
        await self.assertGetTXOK()

    @unittest.skip("Intermittent Failure")
    async def test_getverbosetx(self):
        await self.assertGetVerboseTXOK()

    async def test_gettxs(self):
        await self.assertTxsOK()

    async def test_balance_merkle_proven(self):
        await self.assertBalanceMerkleProvenOK()

    async def test_balances_merkle_proven(self):
        await self.assertBalancesMerkleProvenOK()

    async def test_transaction(self):
        with mock.patch('cryptos.electrumx_client.client.NotificationSession.send_request',
                        side_effect=self.mock_electrumx_send_request):
            await self.assertTransactionOK("9e60e1f5257950837bbbe7f2ca84410a59d9b6326823c4e0defdb01fb115df97")

    async def test_transaction_segwit(self):
        with mock.patch('cryptos.electrumx_client.client.NotificationSession.send_request',
                        side_effect=self.mock_electrumx_send_request):
            await self.assertSegwitTransactionOK(
                "c263194aed2d18f3289105f99636b2187f7228c0a5d3b0c757d68a7a9099a4c4")

    async def test_transaction_native_segwit(self):
        with mock.patch('cryptos.electrumx_client.client.NotificationSession.send_request',
                        side_effect=self.mock_electrumx_send_request):
            await self.assertNativeSegwitTransactionOK(
                "41c45ea2381ec4da825712d1cee6d598247edac93e8fedb79ff6d426bfb2af6f")

    async def test_transaction_mixed_segwit(self):
        with mock.patch('cryptos.electrumx_client.client.NotificationSession.send_request',
                        side_effect=self.mock_electrumx_send_request):
            await self.assertMixedSegwitTransactionOK(
                "5e918bfc41e912b581adb513efe4f480d37950c386842ad1bbd052b3fb5c1a48")

    async def test_transaction_multisig(self):
        with mock.patch('cryptos.electrumx_client.client.NotificationSession.send_request',
                        side_effect=self.mock_electrumx_send_request):
            await self.assertMultiSigTransactionOK(
                "a0c54d9e45a7bd3ab224d154760f08c7fcfc8aeaecbb46914796ff52e27e6c85")

    async def test_transaction_native_segwit_multisig(self):
        with mock.patch('cryptos.electrumx_client.client.NotificationSession.send_request',
                        side_effect=self.mock_electrumx_send_request):
            await self.assertNativeSegwitMultiSigTransactionOK("37d3c7f0ff6846a34bbd69a472cadb83ee631b5ea1bc7695cfe52a3442ed7397")

    async def test_sendmulti_recipient_tx(self):
        with mock.patch('cryptos.electrumx_client.client.NotificationSession.send_request',
                        side_effect=self.mock_electrumx_send_request):
            await self.assertSendMultiRecipientsTXOK("c85e3e899b43c5a1f68647c852ee13f9e3e2b409b4cc73cce131d70dba1a793d")

    async def test_send(self):
        with mock.patch('cryptos.electrumx_client.client.NotificationSession.send_request',
                        side_effect=self.mock_electrumx_send_request):
            await self.assertSendOK("6d2cfbc85d56284f0ee88f977979d20e8e66017b48737568d1b23ae597594e2d")

    async def test_subscribe_block_headers(self):
        await self.assertSubscribeBlockHeadersOK()

    async def test_subscribe_block_headers_sync(self):
        await self.assertSubscribeBlockHeadersSyncCallbackOK()

    async def test_latest_block(self):
        await self.assertLatestBlockOK()

    async def test_confirmations(self):
        await self.assertConfirmationsOK()



from typing import TypedDict, Dict, Any
from typing import List, AnyStr, Union, Callable, Awaitable
from typing_extensions import NotRequired
from .electrumx_client.types import ElectrumXTx


class TxInput(TypedDict):
    tx_hash: str
    tx_pos: int
    script: bytes
    sequence: int
    value: NotRequired[int]
    address: NotRequired[str]


class TxOut(TypedDict):
    value: int
    address: NotRequired[str]
    script: NotRequired[str]


class Witness(TypedDict):
    number: int
    scriptCode: AnyStr


class Tx(TypedDict):
    ins: List[TxInput]
    outs: List[TxOut]
    version: str
    marker: NotRequired[str]
    flag: NotRequired[str]
    witness: NotRequired[List[Witness]]
    tx_hash: NotRequired[str]
    locktime: int


class BlockHeader(TypedDict):
    version: int
    prevhash: bytes
    merkle_root: bytes
    timestamp: int
    bits: int
    nonce: int
    hash: bytes


class MerkleProof(TypedDict):
    tx_hash: str
    siblings: NotRequired[List[str]]
    proven: bool


class AddressBalance(TypedDict):
    address: str
    balance: int


class AddressStatusUpdate(TypedDict):
    address: str
    status: str


BlockHeaderCallbackSync = Callable[[int, str, BlockHeader], None]
BlockHeaderCallbackAsync = Callable[[int, str, BlockHeader], Awaitable[None]]
BlockHeaderCallback = Union[BlockHeaderCallbackSync, BlockHeaderCallbackAsync]


AddressCallbackSync = Callable[[str, str], None]
AddressCallbackAsync = Callable[[str, str], Awaitable[None]]
AddressCallback = Union[AddressCallbackSync, AddressCallbackAsync]


AddressTXCallbackSync = Callable[[str, List[ElectrumXTx], List[ElectrumXTx], List[ElectrumXTx], List[ElectrumXTx], int, int, int], None]
AddressTXCallbackAsync = Callable[[str, List[ElectrumXTx], List[ElectrumXTx], List[ElectrumXTx], List[ElectrumXTx], int, int, int], Awaitable[None]]
AddressTXCallback = Union[AddressTXCallbackSync, AddressTXCallbackAsync]


# Either a single private key or a mapping of addresses to private keys
PrivkeyType = Union[int, str, bytes]
PrivateKeySignAllType = Union[Dict[str, PrivkeyType], PrivkeyType]
PubKeyType = Union[list, tuple, str, bytes]


class TXInspectType(TypedDict):
    ins: Dict[str, TxInput]
    outs: List[Dict[str, Any]]
    fee: int



import unittest

from cryptos import coins_async
from cryptos.main import privtopub, compress
from cryptos.testing.testcases_async import BaseAsyncCoinTestCase
from cryptos.types import ElectrumXTx, TxOut
from cryptos.electrumx_client.types import ElectrumXMultiBalanceResponse
from typing import List


class TestLitecoinTestnet(BaseAsyncCoinTestCase):
    name = "Litecoin Testnet"
    coin = coins_async.Litecoin
    addresses: List[str] = ["n2DQVQZiA2tVBDu4fNAjKVBS2RArWhfncv",
                            "mnjBtsvoSo6dMvMaeyfaCCRV4hAF8WA2cu",
                            "mmbKDFPjBatJmZ6pWTW6yqXSC6826YLBX6"]
    segwit_addresses: List[str] = ["QaRdhvUtEt5vnU1MXUhK8JT1wLqS7Y29Ua",
                                  "QhAx5qBJphxMrSZfwzaf8KyP9T2DrAMbiC",
                                  "QMPRBmeVuqPf8KxYeF9ANdVKh6cNTePk7W"]
    native_segwit_addresses: List[str] = [
        "tltc1q95cgql39zvtc57g4vn8ytzmlvtt43skn39z3vs",
        "tltc1qfuvnn87p787z7nqv9seu4e8fqel83yac3kxh62",
        "tltc1qst3pkm860tjt9y70ugnaluqyqnfa7h54q7xv2n"]
    multisig_addresses: List[str] = ["QQ85DTHTd76GijMxPHduptiToi3KbZFpcw",
                                     "QMossmrfEt4qrduKBuU35988GHPCtUY4ZQ"]
    native_segwit_multisig_addresses: List[str] = [
        "tltc1q7e42a8gmgp5d7kw8myrruj5qnxp8edp7v5y0y95hrjz6t530uehqf3n2q4",
        "tltc1qu7fz4uku8kh6tg7qghj7rnf88g858lal258gzfu85kx7vc5h0qpsm9lgv7"
    ]
    # Private keys for above addresses in same order
    privkeys = ["098ddf01ebb71ead01fc52cb4ad1f5cafffb5f2d052dd233b3cad18e255e1db1",
                "0861e1bb62504f5e9f03b59308005a6f2c12c34df108c6f7c52e5e712a08e91401",
                "c396c62dfdc529645b822dc4eaa7b9ddc97dd8424de09ca19decce61e6732f71"]
    privkey_standard_wifs: List[str] = ['7QMPk7vLsHRiUXSsuJQkVyyWbnqHH3DGHPwo43MMwbRUz7o9yLi',
                                       'VG2jGVfRhLNtN6HHohBYJo9tGXVgJcbyWPjBHPDJrbqRYtCsGKDb',
                                       "7RmKkdCZDPqLCPXoqpqLcJVDi5e4iETUrUAdnBLTAme7nJDsayK"]
    privkey_segwit_wifs: List[str] = ['VZEGXeGF3FLN1XQkmzdcbGfQ9RhQWdJanFX7PL54QvSV5Qcy4Stv',
                                       'VZBzQpPE3eDAK8JpuqyKu9LrmPtyyT9RqGySQHuVqZLr7YZusmsc',
                                       "VfTtusVoY1cbgJUWrev99ANXbBsu3pwLGE9nTKK3V5YFVHKwMvh4"]
    privkey_native_segwit_wifs: List[str] = ["VQedxUuLs6RE31PzDvEio6ZutzVkgD2rcotypsixvSgmo5QjrRhM",
                                             "VQcMqf2KsVJ2LcJ4MmaS6yFNWxhL92shfqMJqqZQM5b8qDNnGqBj",
                                             "VWtGLi8uMrhThnTkJaXFLzH3LkgFDQfc6nXetrxwzbnYCxAXw4MK"]
    fee: int = 2000
    max_fee: int = 4000
    testnet: bool = True

    min_latest_height: int = 2391524
    unspent_addresses: List[str] = ["ms31HApa3jvv3crqvZ3sJj7tC5TCs61GSA"]
    unspent: List[ElectrumXTx] = [{'address': 'ms31HApa3jvv3crqvZ3sJj7tC5TCs61GSA',
                                   'height': 296481,
                                   'tx_hash': '3f7a5460983cdfdf8118a1ab6bc84c28e536e83971532d4910c26bd21153de19',
                                   'tx_pos': 1,
                                   'value': 100000000}]
    unspents: List[ElectrumXTx] = unspent
    balance: ElectrumXMultiBalanceResponse = {'confirmed': 100000000, 'unconfirmed': 0}
    balances: List[ElectrumXMultiBalanceResponse] = [{'address': unspent_addresses[0]} | dict(balance)]
    history: List[ElectrumXTx] = [{'height': 296481, 'tx_hash': '3f7a5460983cdfdf8118a1ab6bc84c28e536e83971532d4910c26bd21153de19'}]
    histories: List[ElectrumXTx] = [{'address': unspent_addresses[0]} | dict(history[0])]
    txid: str = "2a288547460ebe410e98fe63a1900b6452d95ec318efb0d58a5584ac67f27d93"
    txheight: int = 296568
    block_hash: str = "9c557ffb695078e9f79d92b449fc0e61d82c331692258eb38495013aaf636218"
    txinputs: List[TxOut] = []
    raw_tx: str = "010000000384c9a749888ec0b7372a5740dfe8ba059f67d7b48cb1110060a4e666b42ea383000000006b483045022100c7081d2329334a78cde23359da1d9684d60b7fdb3e396c9d2633c419f9ad30ff022058e7cd031df6b7c7208b3140887e9ba012c81e4f300fcf388256f2636b0682e401210391ed6bf1e0842997938ea2706480a7085b8bb253268fd12ea83a68509602b6e0ffffffff84c9a749888ec0b7372a5740dfe8ba059f67d7b48cb1110060a4e666b42ea383010000006a47304402207ceb8ca2179fc4ff975ebc3a95b6b1ddc5ce0c280203576d8a1d53948c7138ac02201157f68003220b7f6c3abc7756e7838e062b81ed511f6caff66aa1a73525efa301210391ed6bf1e0842997938ea2706480a7085b8bb253268fd12ea83a68509602b6e0ffffffff7183a5bf996607a10ee0566716012a587adb9e43778c1a019deb3e43b9537af2010000006b483045022100a95b8b36d08f944949b7fa2dca32f5e44e568339dcde11a8713e4676ed3bc77202204d117c91053b667714b1496583583bf8633b7fb189a800d08fdaaefd3f1ef49301210391ed6bf1e0842997938ea2706480a7085b8bb253268fd12ea83a68509602b6e0ffffffff020cb82d01000000001976a91442a3e11a80b25ff63b2074c51d1745132bccbba188ac74789b0a000000001976a914c384950342cb6f8df55175b48586838b03130fad88ac00000000"

    def test_standard_wif_ok(self):
        self.assertStandardWifOK()

    def test_p2wpkh_p2sh_wif_ok(self):
        self.assertP2WPKH_P2SH_WifOK()

    def test_p2wpkh_wif_ok(self):
        self.assertP2WPKH_WIFOK()

    async def test_balance(self):
        await self.assertBalanceOK()

    async def test_balances(self):
        await self.assertBalancesOK()

    async def test_unspent(self):
        await self.assertUnspentOK()

    async def test_unspents(self):
        await self.assertUnspentsOK()

    async def test_merkle_proof(self):
        await self.assertMerkleProofOK()

    async def test_history(self):
        await self.assertHistoryOK()

    async def test_histories(self):
        await self.assertHistoriesOK()

    async def test_block_header(self):
        await self.assertBlockHeaderOK()

    async def test_block_headers(self):
        await self.assertBlockHeadersOK()

    async def test_gettx(self):
        await self.assertGetTXOK()

    @unittest.skip("Intermittent Failure")
    async def test_getverbosetx(self):
        await self.assertGetVerboseTXOK()

    async def test_gettxs(self):
        await self.assertGetTXOK()

    async def test_balance_merkle_proven(self):
        await self.assertBalanceMerkleProvenOK()

    async def test_balances_merkle_proven(self):
        await self.assertBalancesMerkleProvenOK()

    @unittest.skip("Intermittent Failure")
    async def test_transaction(self):
        """
        Sample transaction:
        TxID: 319107640d6dbbbf0da7c6341074410862f61e167c49094477e10f80ecf41b20
        """
        await self.assertTransactionOK()

    async def test_transaction_segwit(self):
        """
        Sample transaction:
        TxID: cc1aba794b8f7bb0176906e724594ce6c79f598cd904c1483de4b43331b25cc5
        """
        await self.assertSegwitTransactionOK()

    @unittest.skip("Intermittent Failure")
    async def test_transaction_native_segwit(self):
        """
        Sample transaction:
        TxID: 8dbaffbd6206147389e0bb9c30c84e9f0a5dedb9a1e15563c950c7470b3c90c5
        """
        await self.assertNativeSegwitTransactionOK()

    async def test_transaction_mixed_segwit(self):
        """
        Sample transaction:
        TxID: 3dc675d5addfad9d142f51411372ed2ad9700a19019114aab775048effadf7da
        """
        await self.assertMixedSegwitTransactionOK()

    async def test_transaction_multisig(self):
        """
        Sample transaction:
        TxID: 6d68cdc662fc92164b371aba385aec9afc99058a6ce6dfaa5c5766705e099f85
        """
        await self.assertMultiSigTransactionOK()

    @unittest.skip("Intermittent Failure")
    async def test_transaction_native_segwit_multisig(self):
        """
        Sample transaction:
        TxID: 4076340f359668b325808ec98888863402c0fb829f0f1d15e2e8ab816841d749
        """
        await self.assertNativeSegwitMultiSigTransactionOK()

    @unittest.skip("Intermittent Failure")
    async def test_sendmulti_recipient_tx(self):
        """
        Sample transaction:
        TxID: da7c7f99f2f1968bfd53dc7a817b4a9f3f3f596727194c4df57fac3f0e68e777
        """
        await self.assertSendMultiRecipientsTXOK()

    async def test_send(self):
        """
        Sample transaction:
        TxID: 779221217a57596149d1ea4bf017066df6a488d1afe730d8d805bbaa6b6b0deb
        """
        await self.assertSendOK()

    async def test_subscribe_block_headers(self):
        await self.assertSubscribeBlockHeadersOK()

    async def test_subscribe_block_headers_sync(self):
        await self.assertSubscribeBlockHeadersSyncCallbackOK()

    async def test_latest_block(self):
        await self.assertLatestBlockOK()

    async def test_confirmations(self):
        await self.assertConfirmationsOK()

    async def test_subscribe_address(self):
        await self.assertSubscribeAddressOK()

    async def test_subscribe_address_sync(self):
        await self.assertSubscribeAddressSyncCallbackOK()

    async def test_subscribe_address_transactions(self):
        await self.assertSubscribeAddressTransactionsOK()

    async def test_subscribe_address_transactions_sync(self):
        await self.assertSubscribeAddressTransactionsSyncOK()



import unittest

from cryptos import coins_async
from cryptos.testing.testcases_async import BaseAsyncCoinTestCase
from typing import List


@unittest.skip('Only Doge Electrumx server is often down')
class TestDogeTestnet(BaseAsyncCoinTestCase):
    name = "Doge Testnet"
    coin = coins_async.Doge
    addresses = ['nptcTdAHaPpEp6BEiCqNHjj1HRgjtFELjM',
                 'nbQPs6XNsA2NzndkhpLDASy4Khg8ZfhUfj',
                 'naGXBTzJbwp4QRNzZJAjx651T6duZy2kgV']
    multisig_addresses: List[str] = ["2MvmK6SRDc13BaYbumBbtkCH2fKbViC5XEv",
                                     "2MtT7kkzRDn1kiT9GZoS1zSgh7twP145Qif"]
    privkeys: List[str] = ["098ddf01ebb71ead01fc52cb4ad1f5cafffb5f2d052dd233b3cad18e255e1db1",
                           "0861e1bb62504f5e9f03b59308005a6f2c12c34df108c6f7c52e5e712a08e91401",
                           "c396c62dfdc529645b822dc4eaa7b9ddc97dd8424de09ca19decce61e6732f71"]
    privkey_standard_wifs: List[str] = ['95YcZiMwUZF3DYW9EHSiC1xEPVdBnFM3dMVBFXkBeMksf8k8F53',
                                       'cf2FrZgQT2Bm5xwhTUvC7VtEiFYJ3YAKeUshG6Y3QXX1dSAZ9s9h',
                                       '96xYaDe9pfeewQb5AosJJLTwVnRyDSbGCRi1yfjGsXyWTNCJpxv']
    fee: int = 300000
    max_fee: int = 1000000
    testnet = True
    min_latest_height = 4109697

    unspent_addresses = ["ncst7MzasMBQFwx2LuCwj8Z6F4pCRppzBP"]
    unspent = [{'address': 'ncst7MzasMBQFwx2LuCwj8Z6F4pCRppzBP',
                'height': 4107911,
                'tx_hash': '268bff71633310c974e262a67a057a618f6d6596bfa95e45692d6c9269a13873',
                'tx_pos': 0,
                'value': 10000000000}]
    unspents = unspent
    balance = {'confirmed': 10000000000, 'unconfirmed': 0}
    balances = [{'address': unspent_addresses[0]} | dict(balance)]
    history = [{'height': 4107911,
                'tx_hash': '268bff71633310c974e262a67a057a618f6d6596bfa95e45692d6c9269a13873'}]
    histories = [{'address': unspent_addresses[0]} | dict(history[0])]
    txid: str = "268bff71633310c974e262a67a057a618f6d6596bfa95e45692d6c9269a13873"
    txheight: int = 4107911
    block_hash: str = "de125596a7937dcccdb49bb55ea50b776fbe32f8bfa9afcda236dcdc6b9880e3"
    raw_tx: str = "0100000001502100e4c8a3c0bb1aabc6698b128ffdaf06be518f126fe3400880920d680600010000006b483045022100918045061097ed4e3ae63d3f5a0fa5aa188a0e1f8f0041fb0693a0c05cb3ce8b02206efcb89204e4924a71704e105b1d8c7cc8a701a70262efb91dc426cdb97c40670121030d78465ee4506029a720bd691c883705391271d9a8f428abae5c577ce44841f1feffffff0200e40b54020000001976a9145f44291007a8a4000b9b8424843793f402638e9388acaae442dc310000001976a914a7106a8a40e93b097a383446d30ce8d1f7af1db288ac86ae3e00"

    def test_standard_wif_ok(self):
        self.assertStandardWifOK()

    async def test_balance(self):
        await self.assertBalanceOK()

    async def test_balances(self):
        await self.assertBalancesOK()

    async def test_unspent(self):
        await self.assertUnspentOK()

    async def test_unspents(self):
        await self.assertUnspentsOK()

    async def test_merkle_proof(self):
        await self.assertMerkleProofOK()

    async def test_history(self):
        await self.assertHistoryOK()

    async def test_histories(self):
        await self.assertHistoriesOK()

    async def test_balance_merkle_proven(self):
        await self.assertBalanceMerkleProvenOK()

    async def test_balances_merkle_proven(self):
        await self.assertBalancesMerkleProvenOK()

    async def test_block_header(self):
        await self.assertBlockHeaderOK()

    async def test_block_headers(self):
        await self.assertBlockHeadersOK()

    async def test_gettx(self):
        await self.assertGetTXOK()

    async def test_getverbosetx(self):
        await self.assertGetVerboseTXOK()

    async def test_gettxs(self):
        await self.assertGetTxsOK()

    async def test_transaction(self):
        """
        Sample transaction:
        TxID: 4d859f797f678f0cbe608490e0ba5bc7f1fa670b34192bbbf6f468fb2ed00fe4
        """
        await self.assertTransactionOK()

    async def test_transaction_multisig(self):
        """
        Sample transaction:
        TxID: 092df25d3bb39bb82e1d00faf08a0aec2255a3214bab2a1a7c6437dd77b6281b
        """
        await self.assertMultiSigTransactionOK()

    @unittest.skip("Intermittent failure")
    async def test_sendmulti_recipient_tx(self):
        """
        Sample transaction:
        TxID: 5723b84848f00d2b315086ce564e535bd6d75fa13225ff75fce0d9ed7cd306a9
        """
        await self.assertSendMultiRecipientsTXOK()

    async def test_send(self):
        """
        Sample transaction:
        TxID: 21fdea144f28d0cb1da99c5fe7c96268aaa98ddef3a14fd627b44ea31ce0be3e
        """
        await self.assertSendOK()

    async def test_subscribe_block_headers(self):
        await self.assertSubscribeBlockHeadersOK()

    async def test_subscribe_block_headers_sync(self):
        await self.assertSubscribeBlockHeadersSyncCallbackOK()

    async def test_latest_block(self):
        await self.assertLatestBlockOK()

    async def test_confirmations(self):
        await self.assertConfirmationsOK()

    async def test_subscribe_address(self):
        await self.assertSubscribeAddressOK()

    async def test_subscribe_address_sync(self):
        await self.assertSubscribeAddressSyncCallbackOK()

    async def test_subscribe_address_transactions(self):
        await self.assertSubscribeAddressTransactionsOK()

    async def test_subscribe_address_transactions_sync(self):
        await self.assertSubscribeAddressTransactionsSyncOK()




from .main import *
from .transaction import deserialize
from .types import BlockHeader, MerkleProof
from binascii import hexlify


def serialize_header(inp: BlockHeader) -> bytes:
    o = encode(inp['version'], 256, 4)[::-1] + \
        inp['prevhash'].decode('hex')[::-1] + \
        inp['merkle_root'].decode('hex')[::-1] + \
        encode(inp['timestamp'], 256, 4)[::-1] + \
        encode(inp['bits'], 256, 4)[::-1] + \
        encode(inp['nonce'], 256, 4)[::-1]
    h = bin_sha256(bin_sha256(o))[::-1].encode('hex')
    assert h == inp['hash'], (sha256(o), inp['hash'])
    return o.encode('hex')


def deserialize_header(inp: bytes) -> BlockHeader:
    return {
        "version": decode(inp[:4][::-1], 256),
        "prevhash": hexlify(inp[4:36][::-1]),
        "merkle_root": hexlify(inp[36:68][::-1]),
        "timestamp": decode(inp[68:72][::-1], 256),
        "bits": decode(inp[72:76][::-1], 256),
        "nonce": decode(inp[76:80][::-1], 256),
        "hash": bin_sha256(bin_sha256(inp))[::-1]
    }


def mk_merkle_proof(merkle_root: bytes, hashes: List[str], index: int) -> MerkleProof:
    """
    This function requires all transaction hashes in a block to be provided
    """
    tx_hash = hashes[index]
    try:
        nodes = [safe_from_hex(h)[::-1] for h in hashes]
        if len(nodes) % 2 and len(nodes) > 2:
            nodes.append(nodes[-1])
        layers = [nodes]
        while len(nodes) > 1:
            newnodes = []
            for i in range(0, len(nodes) - 1, 2):
                newnodes.append(bin_sha256(bin_sha256(nodes[i] + nodes[i+1])))
            if len(newnodes) % 2 and len(newnodes) > 2:
                newnodes.append(newnodes[-1])
            nodes = newnodes
            layers.append(nodes)
        # Sanity check, make sure merkle root is valid
        assert bytes_to_hex_string(nodes[0][::-1]) == merkle_root
        merkle_siblings = \
            [layers[i][(index >> i) ^ 1] for i in range(len(layers)-1)]
        return {
            "tx_hash": tx_hash,
            "siblings": [bytes_to_hex_string(x[::-1]) for x in merkle_siblings],
            'proven': True
        }
    except:
        return {
            "tx_hash": tx_hash,
            "siblings": [],
            'proven': False
        }


def verify_merkle_proof(tx_hash: str, merkle_root: bytes, hashes: List[str], index: int) -> MerkleProof:
    h = safe_from_hex(tx_hash)[::-1]
    nodes = [safe_from_hex(h)[::-1] for h in hashes]
    proven = True
    for item in nodes:
        inner_node = (item + h) if (index & 1) else (h + item)
        try:
            deserialize(inner_node)
        except Exception as e:
            pass
        else:
            proven = False          # If a node serializes as a transaction, could be an attack
            break
        h = bin_sha256(bin_sha256(inner_node))
        index >>= 1
    if index != 0:
        proven = False
    h = bytes_to_hex_string(h[::-1]).encode()
    return {
            "tx_hash": tx_hash,
            'proven': proven and h == merkle_root
    }
